from datetime import datetime

import pylab as plt
import torch

from ai_traineree.agents.td3 import TD3Agent as Agent
from ai_traineree.loggers.tensorboard_logger import TensorboardLogger
from ai_traineree.runners.env_runner import EnvRunner
from ai_traineree.tasks import GymTask
from ai_traineree.types import TaskType

seed = 32167
torch.manual_seed(seed)


env_name = "LunarLanderContinuous-v2"
task: TaskType = GymTask(env_name, seed=seed)
config = {
    "warm_up": 1000,
    "batch_size": 100,
    "hidden_layers": (100, 100),
    "number_updates": 3,
    "actor_lr": 1e-3,
    "critic_lr": 1e-3,
    "update_freq": 20,
    "update_policy_freq": 40,
    "noise_scale": 1.0,
}
agent = Agent(task.obs_space, task.action_space, **config)

log_dir = f"runs/{env_name}_{agent.model}-{datetime.now().isoformat()[:-7]}"
data_logger = TensorboardLogger(log_dir=log_dir)
env_runner = EnvRunner(task, agent, data_logger=data_logger, window_len=30, debug_log=True)

# interact_episode(task, agent, 0, render=True)
scores = env_runner.run(reward_goal=80, max_episodes=2000, eps_end=0.2, eps_decay=0.99, force_new=True)
# env_runner.interact_episode(eps=0, render=True)

data_logger.close()

# plot the scores
fig = plt.figure()
ax = fig.add_subplot(111)
plt.plot(range(len(scores)), scores)
plt.ylabel("Score")
plt.xlabel("Episode #")
plt.savefig(f"{env_name}.png", dpi=120)
plt.show()



import copy
from typing import Callable, Dict, List, Optional

import torch
import torch.nn as nn
import torch.optim as optim

from ai_traineree import DEVICE
from ai_traineree.agents import AgentBase
from ai_traineree.agents.agent_utils import soft_update
from ai_traineree.buffers.buffer_factory import BufferFactory
from ai_traineree.buffers.nstep import NStepBuffer
from ai_traineree.buffers.per import PERBuffer
from ai_traineree.loggers import DataLogger
from ai_traineree.networks.heads import RainbowNet
from ai_traineree.types import AgentState, BufferState, NetworkState
from ai_traineree.types.dataspace import DataSpace
from ai_traineree.types.experience import Experience
from ai_traineree.utils import to_numbers_seq, to_tensor


class RainbowAgent(AgentBase):
    """Rainbow agent as described in [1].

    Rainbow is a DQN agent with some improvements that were suggested before 2017.
    As mentioned by the authors it's not exhaustive improvement but all changes are in
    relatively separate areas so their connection makes sense. These improvements are:
    * Priority Experience Replay
    * Multi-step
    * Double Q net
    * Dueling nets
    * NoisyNet
    * CategoricalNet for Q estimate

    Consider this class as a particular version of the DQN agent.

    [1] "Rainbow: Combining Improvements in Deep Reinforcement Learning" by Hessel et al. (DeepMind team)
        https://arxiv.org/abs/1710.02298

    """

    model = "Rainbow"

    def __init__(
        self,
        obs_space: DataSpace,
        action_space: DataSpace,
        state_transform: Optional[Callable] = None,
        reward_transform: Optional[Callable] = None,
        **kwargs,
    ):
        """
        A wrapper over the DQN thus majority of the logic is in the DQNAgent.
        Special treatment is required because the Rainbow agent uses categorical nets
        which operate on probability distributions. Each action is taken as the estimate
        from such distributions.

        Parameters:
            obs_space (DataSpace): Dataspace describing the input.
            action_space (DataSpace): Dataspace describing the output.
            state_transform (optional func):
            reward_transform (optional func):

        Keyword arguments:
            pre_network_fn (function that takes input_shape and returns network):
                Used to preprocess state before it is used in the value- and advantage-function in the dueling nets.
            hidden_layers (tuple of ints): Shape of the hidden layers in fully connected network. Default: (100, 100).
            lr (default: 1e-3): Learning rate value.
            gamma (float): Discount factor. Default: 0.99.
            tau (float): Soft-copy factor. Default: 0.002.
            update_freq (int): Number of steps between each learning step. Default 1.
            batch_size (int): Number of samples to use at each learning step. Default: 80.
            buffer_size (int): Number of most recent samples to keep in memory for learning. Default: 1e5.
            warm_up (int): Number of samples to observe before starting any learning step. Default: 0.
            number_updates (int): How many times to use learning step in the learning phase. Default: 1.
            max_grad_norm (float): Maximum norm of the gradient used in learning. Default: 10.
            using_double_q (bool): Whether to use Double Q Learning network. Default: True.
            n_steps (int): Number of lookahead steps when estimating reward. See :class:`NStepBuffer`. Default: 3.
            v_min (float): Lower bound for distributional value V. Default: -10.
            v_max (float): Upper bound for distributional value V. Default: 10.
            num_atoms (int): Number of atoms (discrete states) in the value V distribution. Default: 21.

        """
        super().__init__(**kwargs)
        self.device = self._register_param(kwargs, "device", DEVICE, update=True)

        self.obs_space = obs_space
        self.action_space = action_space
        self._config["obs_space"] = self.obs_space
        self._config["action_space"] = self.action_space
        self.action_size = action_space.to_feature()

        self.lr = float(self._register_param(kwargs, "lr", 3e-4))
        self.gamma = float(self._register_param(kwargs, "gamma", 0.99))
        self.tau = float(self._register_param(kwargs, "tau", 0.002))
        self.update_freq = int(self._register_param(kwargs, "update_freq", 1))
        self.batch_size = int(self._register_param(kwargs, "batch_size", 80, update=True))
        self.buffer_size = int(self._register_param(kwargs, "buffer_size", int(1e5), update=True))
        self.warm_up = int(self._register_param(kwargs, "warm_up", 0))
        self.number_updates = int(self._register_param(kwargs, "number_updates", 1))
        self.max_grad_norm = float(self._register_param(kwargs, "max_grad_norm", 10))

        self.iteration: int = 0
        self.using_double_q = bool(self._register_param(kwargs, "using_double_q", True))

        self.state_transform = state_transform if state_transform is not None else lambda x: x
        self.reward_transform = reward_transform if reward_transform is not None else lambda x: x

        v_min = float(self._register_param(kwargs, "v_min", -10))
        v_max = float(self._register_param(kwargs, "v_max", 10))
        self.num_atoms = int(self._register_param(kwargs, "num_atoms", 21, drop=True))
        self.z_atoms = torch.linspace(v_min, v_max, self.num_atoms, device=self.device)
        self.z_delta = self.z_atoms[1] - self.z_atoms[0]

        self.buffer = PERBuffer(**kwargs)
        self.__batch_indices = torch.arange(self.batch_size, device=self.device)

        self.n_steps = int(self._register_param(kwargs, "n_steps", 3))
        self.n_buffer = NStepBuffer(n_steps=self.n_steps, gamma=self.gamma)

        # Note that in case a pre_network is provided, e.g. a shared net that extracts pixels values,
        # it should be explicitly passed in kwargs
        kwargs["hidden_layers"] = to_numbers_seq(self._register_param(kwargs, "hidden_layers", (100, 100)))
        self.net = RainbowNet(obs_space.shape, self.action_size, num_atoms=self.num_atoms, **kwargs)
        self.target_net = RainbowNet(obs_space.shape, self.action_size, num_atoms=self.num_atoms, **kwargs)

        self.optimizer = optim.Adam(self.net.parameters(), lr=self.lr)
        self.dist_probs = None
        self._loss = float("nan")

    @property
    def loss(self):
        return {"loss": self._loss}

    @loss.setter
    def loss(self, value):
        if isinstance(value, dict):
            value = value["loss"]
        self._loss = value

    def step(self, experience: Experience) -> None:
        """Letting the agent to take a step.

        On some steps the agent will initiate learning step. This is dependent on
        the `update_freq` value.

        Parameters:
            obs (ObservationType): Observation.
            action (int): Discrete action associated with observation.
            reward (float): Reward obtained for taking action at state.
            next_obs (ObservationType): Observation in a state where the action took.
            done: (bool) Whether in terminal (end of episode) state.

        """
        assert isinstance(experience.action, int), "Rainbow expects discrete action (int)"
        self.iteration += 1
        t_obs = to_tensor(self.state_transform(experience.obs)).float().to("cpu")
        t_next_obs = to_tensor(self.state_transform(experience.next_obs)).float().to("cpu")
        reward = self.reward_transform(experience.reward)

        # Delay adding to buffer to account for n_steps (particularly the reward)
        self.n_buffer.add(
            obs=t_obs.numpy(),
            action=[int(experience.action)],
            reward=[reward],
            done=[experience.done],
            next_obs=t_next_obs.numpy(),
        )
        if not self.n_buffer.available:
            return

        self.buffer.add(**self.n_buffer.get().get_dict())

        if self.iteration < self.warm_up:
            return

        if len(self.buffer) >= self.batch_size and (self.iteration % self.update_freq) == 0:
            for _ in range(self.number_updates):
                self.learn(self.buffer.sample())

            # Update networks only once - sync local & target
            soft_update(self.target_net, self.net, self.tau)

    def act(self, experience: Experience, eps: float = 0.0) -> Experience:
        """
        Returns actions for given state as per current policy.

        Parameters:
            state: Current available state from the environment.
            epislon: Epsilon value in the epislon-greedy policy.

        """
        # Epsilon-greedy action selection
        if self._rng.random() < eps:
            # TODO: Update with action_space.sample() once implemented
            assert len(self.action_space.shape) == 1, "Only 1D is supported right now"
            action = self._rng.randint(self.action_space.low, self.action_space.high)
            return experience.update(action=action)

        t_obs = to_tensor(self.state_transform(experience.obs)).float().unsqueeze(0).to(self.device)
        self.dist_probs = self.net.act(t_obs)
        q_values = (self.dist_probs * self.z_atoms).sum(-1)
        action = int(q_values.argmax(-1))  # Action maximizes state-action value Q(s, a)
        return experience.update(action=action)

    def learn(self, experiences: Dict[str, List]) -> None:
        """
        Parameters:
            experiences: Contains all experiences for the agent. Typically sampled from the memory buffer.
                Five keys are expected, i.e. `state`, `action`, `reward`, `next_state`, `done`.
                Each key contains a array and all arrays have to have the same length.

        """
        rewards = to_tensor(experiences["reward"]).float().to(self.device)
        dones = to_tensor(experiences["done"]).type(torch.int).to(self.device)
        obss = to_tensor(experiences["obs"]).float().to(self.device)
        next_obss = to_tensor(experiences["next_obs"]).float().to(self.device)
        actions = to_tensor(experiences["action"]).type(torch.long).to(self.device)
        assert rewards.shape == dones.shape == (self.batch_size, 1)
        assert obss.shape == next_obss.shape == (self.batch_size,) + self.obs_space.shape
        assert actions.shape == (self.batch_size, 1)  # Discrete domain

        with torch.no_grad():
            prob_next = self.target_net.act(next_obss)
            q_next = (prob_next * self.z_atoms).sum(-1) * self.z_delta
            if self.using_double_q:
                duel_prob_next = self.net.act(next_obss)
                a_next = torch.argmax((duel_prob_next * self.z_atoms).sum(-1), dim=-1)
            else:
                a_next = torch.argmax(q_next, dim=-1)

            prob_next = prob_next[self.__batch_indices, a_next, :]

        m = self.net.dist_projection(rewards, 1 - dones, self.gamma**self.n_steps, prob_next)
        assert m.shape == (self.batch_size, self.num_atoms)

        log_prob = self.net(obss, log_prob=True)
        assert log_prob.shape == (self.batch_size,) + self.action_size + (self.num_atoms,)
        log_prob = log_prob[self.__batch_indices, actions.squeeze(), :]
        assert log_prob.shape == m.shape == (self.batch_size, self.num_atoms)

        # Cross-entropy loss error and the loss is batch mean
        error = -torch.sum(m * log_prob, 1)
        assert error.shape == (self.batch_size,)
        loss = error.mean()
        assert loss >= 0

        self.optimizer.zero_grad()
        loss.backward()
        nn.utils.clip_grad_norm_(self.net.parameters(), self.max_grad_norm)
        self.optimizer.step()
        self._loss = float(loss.item())

        if hasattr(self.buffer, "priority_update"):
            assert (~torch.isnan(error)).any()
            self.buffer.priority_update(experiences["index"], error.detach().cpu().numpy())

        # Update networks - sync local & target
        soft_update(self.target_net, self.net, self.tau)

    def state_dict(self) -> Dict[str, dict]:
        """Returns agent's state dictionary.

        Returns:
            State dicrionary for internal networks.

        """
        return {"net": self.net.state_dict(), "target_net": self.target_net.state_dict()}

    def log_metrics(self, data_logger: DataLogger, step: int, full_log: bool = False):
        data_logger.log_value("loss/agent", self._loss, step)

        if full_log and self.dist_probs is not None:
            assert len(self.action_space.shape) == 1, "Only 1D actions currently supported"
            action_size = self.action_size[0]
            for action_idx in range(action_size):
                dist = self.dist_probs[0, action_idx]
                data_logger.log_value(f"dist/expected_{action_idx}", (dist * self.z_atoms).sum().item(), step)
                data_logger.add_histogram(
                    f"dist/Q_{action_idx}",
                    min=self.z_atoms[0],
                    max=self.z_atoms[-1],
                    num=len(self.z_atoms),
                    sum=dist.sum(),
                    sum_squares=dist.pow(2).sum(),
                    bucket_limits=self.z_atoms + self.z_delta,
                    bucket_counts=dist,
                    global_step=step,
                )

        # This method, `log_metrics`, isn't executed on every iteration but just in case we delay plotting weights.
        # It simply might be quite costly. Thread wisely.
        if full_log:
            for idx, layer in enumerate(self.net.value_net.layers):
                if hasattr(layer, "weight"):
                    data_logger.create_histogram(f"value_net/layer_weights_{idx}", layer.weight.cpu(), step)
                if hasattr(layer, "bias") and layer.bias is not None:
                    data_logger.create_histogram(f"value_net/layer_bias_{idx}", layer.bias.cpu(), step)
            for idx, layer in enumerate(self.net.advantage_net.layers):
                if hasattr(layer, "weight"):
                    data_logger.create_histogram(f"advantage_net/layer_{idx}", layer.weight.cpu(), step)
                if hasattr(layer, "bias") and layer.bias is not None:
                    data_logger.create_histogram(f"advantage_net/layer_bias_{idx}", layer.bias.cpu(), step)

    def get_state(self) -> AgentState:
        """Provides agent's internal state."""
        return AgentState(
            model=self.model,
            obs_space=self.obs_space,
            action_space=self.action_space,
            config=self._config,
            buffer=copy.deepcopy(self.buffer.get_state()),
            network=copy.deepcopy(self.get_network_state()),
        )

    def get_network_state(self) -> NetworkState:
        return NetworkState(net=dict(net=self.net.state_dict(), target_net=self.target_net.state_dict()))

    @staticmethod
    def from_state(state: AgentState) -> AgentBase:
        config = copy.copy(state.config)
        config.update({"obs_space": state.obs_space, "action_space": state.action_space})
        agent = RainbowAgent(**config)
        if state.network is not None:
            agent.set_network(state.network)
        if state.buffer is not None:
            agent.set_buffer(state.buffer)
        return agent

    def set_network(self, network_state: NetworkState) -> None:
        self.net.load_state_dict(network_state.net["net"])
        self.target_net.load_state_dict(network_state.net["target_net"])

    def set_buffer(self, buffer_state: BufferState) -> None:
        self.buffer = BufferFactory.from_state(buffer_state)

    def save_state(self, path: str) -> None:
        """Saves agent's state into a file.

        Parameters:
            path: String path where to write the state.

        """
        agent_state = self.get_state()
        torch.save(agent_state, path)

    def load_state(self, path: str) -> None:
        """Loads state from a file under provided path.

        Parameters:
            path: String path indicating where the state is stored.

        """
        agent_state = torch.load(path)
        self._config = agent_state.get("config", {})
        self.__dict__.update(**self._config)

        self.net.load_state_dict(agent_state["net"])
        self.target_net.load_state_dict(agent_state["target_net"])

    def save_buffer(self, path: str) -> None:
        """Saves data from the buffer into a file under provided path.

        Parameters:
            path: String path where to write the buffer.

        """
        import json

        dump = self.buffer.dump_buffer(serialize=True)
        with open(path, "w") as f:
            json.dump(dump, f)

    def load_buffer(self, path: str) -> None:
        """Loads data into the buffer from provided file path.

        Parameters:
            path: String path indicating where the buffer is stored.

        """
        import json

        with open(path, "r") as f:
            buffer_dump = json.load(f)
        self.buffer.load_buffer(buffer_dump)

    def __eq__(self, o: object) -> bool:
        return (
            super().__eq__(o)
            and isinstance(o, type(self))
            and self._config == o._config
            and self.buffer == o.buffer
            and self.get_network_state() == o.get_network_state()
        )



import numpy as np


def generate_sample_SARS(iterations, obs_size: int = 4, action_size: int = 2, dict_type=False):
    def state_fn():
        return np.random.random(obs_size)

    def action_fn():
        return np.random.random(action_size)

    def reward_fn():
        return float(np.random.random() - 0.5)

    def done_fn():
        return np.random.random() > 0.5

    state = state_fn()

    for _ in range(iterations):
        next_state = state_fn()
        if dict_type:
            yield dict(
                state=list(state),
                action=list(action_fn()),
                reward=[reward_fn()],
                next_state=list(next_state),
                done=[bool(done_fn())],
            )
        else:
            yield (list(state), list(action_fn()), reward_fn(), list(next_state), bool(done_fn()))
        state = next_state



import abc
from typing import Dict


class DataLogger(abc.ABC):
    def __dell__(self):
        self.close()

    @abc.abstractmethod
    def close(self) -> None:
        ...

    @abc.abstractmethod
    def set_hparams(self, *, hparams: Dict) -> None:
        ...

    @abc.abstractmethod
    def log_value(self, name, value, step) -> None:
        ...

    @abc.abstractmethod
    def log_values_dict(self, name, values, step) -> None:
        ...

    @abc.abstractmethod
    def add_histogram(self, *args, **kwargs) -> None:
        ...

    @abc.abstractmethod
    def create_histogram(self, name, values, step) -> None:
        """Creates a histogram out of provided data."""
        ...



from functools import lru_cache
from typing import Optional, Tuple

import torch
import torch.nn as nn
from torch.distributions import Beta, Dirichlet, MultivariateNormal, Normal
from torch.distributions.distribution import Distribution

from ai_traineree.networks import NetworkType
from ai_traineree.networks.bodies import FcNet
from ai_traineree.types import FeatureType


class PolicyType(NetworkType):
    param_dim: int


class MultivariateGaussianPolicySimple(PolicyType):
    """
    Multivariate Gaussian (Normal) Policy.

    Simplicity of this class, compared to `MultivariateGaussianPolicy`, is in
    the assumption that the covariance is diagonal, sample independent and
    is treated a trainable parameter.
    """

    param_dim = 1

    def __init__(
        self, size: int, std_init: float = 0.5, std_min: float = 0.0001, std_max: float = 2.0, device=None, **kwargs
    ):
        """
        Parameters:
            size (int): Size of the observation.
            std_init: Initial value for covariance's diagonal. All values start the same. Default: 0.5.
            std_min: Minimum value for standard deviation. Default: 0.0001.
            std_max: Maximum value for standard deviation. Default: 2.
            device: Device where to allocate memory. CPU or CUDA.

        """
        super(MultivariateGaussianPolicySimple, self).__init__()
        self.size = size
        self.dist = Normal if size == 1 else MultivariateNormal
        self.std_min = std_min
        self.std_max = std_max
        std_init = float(max(min(std_max, std_init), std_min))
        self.std = nn.Parameter(torch.full((self.size,), std_init, device=device))

    @staticmethod
    @lru_cache(maxsize=10)
    def _empty_std(batch_size: int, size: int, device):
        return torch.zeros((batch_size, size, size), device=device)

    @staticmethod
    @lru_cache(maxsize=10)
    def diag_idx(batch_size: int, size: int, device):
        return torch.arange(size, device=device).repeat((batch_size, 1, 1))

    def forward(self, x, deterministic: bool = False) -> Distribution:
        """Samples from distribution.

        Parameters:
            x (tensor): Uses a location (mu) for the distrubition.
            deterministic (bool): Whether to sample from distribution, or use estimates.
                Default: False, i.e. it'll sample distribution.

        """
        if deterministic:
            return x.view(-1, 1)

        x = x.float()  # Need to convert as dist doesn't work on int
        self.std.data.clamp_(self.std_min, self.std_max)

        if self.size == 1:
            return self.dist(x.view(-1, 1), scale=self.std.view(-1, 1))

        batch_size = x.shape[0]
        new_shape = (batch_size, 1, 1)
        idx = self.diag_idx(batch_size, self.size, device=x.device)
        std_shell = self._empty_std(batch_size, self.size, x.device)
        std_vals = self.std.repeat(new_shape)
        std = std_shell.scatter(1, idx, std_vals)
        self._last_samples = {"mu": x, "std": self.std}
        self._last_dist = self.dist(x, scale_tril=std)
        return self._last_dist.rsample()

    def act(self, x):
        return self.forward(x, deterministic=True)

    def log_prob(self, samples) -> torch.Tensor:
        assert self._last_dist is not None, "Need to execute `forward` on data first"
        return self._last_dist.log_prob(samples)


class MultivariateGaussianPolicy(PolicyType):
    """
    Multivariate Gaussian (Normal) Policy.

    In contrast to `MultivariateGaussianPolicySimple` it assumes that
    distribution's characteristics are estimated by the network rather
    than optimized by the optimizer.
    Both location and covariance are assumed to be inputs into the policy.

    """

    param_dim = 2

    def __init__(
        self, size: int, std_init: float = 1.0, std_min: float = 0.001, std_max: float = 2.0, device=None, **kwargs
    ):
        """
        Parameters:
            size: Observation's dimensionality upon sampling.
            batch_size: Expected size of batch.
            device: Device where to allocate memory. CPU or CUDA.
        """
        super(MultivariateGaussianPolicy, self).__init__()
        self.size = size
        self.dist = Normal if size == 1 else MultivariateNormal
        self.std_init = std_init
        self.std_min = std_min
        self.std_max = std_max
        self._last_samples = None

    @staticmethod
    @lru_cache(maxsize=10)
    def _empty_std(batch_size: int, size: int, device):
        return torch.zeros((batch_size, size, size), device=device)

    @staticmethod
    @lru_cache(maxsize=10)
    def diag_idx(batch_size: int, size: int, device):
        return torch.arange(size, device=device).repeat((batch_size, 1, 1)).view(batch_size, size, 1)

    def forward(self, x, deterministic=False) -> Distribution:
        """Returns distribution"""
        x = x.view(-1, self.param_dim, self.size)
        mu = x[:, 0]
        if deterministic:
            return mu.view(-1, 1)

        std = torch.clamp(x[:, 1], self.std_min, self.std_max).unsqueeze(-1)

        if self.size == 1:
            _mu = mu.view(-1, 1)
            _scale = std.view(-1, 1)
            self._last_samples = {"mu": _mu, "std": torch.diagonal(_scale)}
            self._last_dist = self.dist(mu.view(-1, 1), scale=_scale)
            return self._last_dist.rsample()

        batch_size = x.shape[0]
        std_shell = self._empty_std(batch_size, self.size, x.device)
        idx = self.diag_idx(batch_size, self.size, x.device)
        std = std_shell.scatter(-1, idx, std)
        _std_values = torch.diagonal(std.squeeze(0))
        self._last_samples = {"mu": mu.squeeze(0), "std": _std_values}
        self._last_dist = self.dist(mu.squeeze(0), std.squeeze(0))

        return self._last_dist.rsample()

    def act(self, x) -> torch.Tensor:
        """Deterministic pass. Ignores covariance and returns locations directly."""
        return x.view(-1, self.size, self.param_dim)[..., 0]

    def log_prob(self, samples):
        return self._last_dist.log_prob(samples)


class GaussianPolicy(PolicyType):
    """
    Univariate Gaussian (Normal) Distribution.
    Has two heads; one for location estimate and one for standard deviation.
    """

    def __init__(self, in_features: FeatureType, out_features: FeatureType, out_scale: float = 1, **kwargs):
        """
        Parameters:
            size: Observation's dimensionality upon sampling.

        """
        super(GaussianPolicy, self).__init__()

        self.in_features = in_features
        self.out_features = out_features
        self.out_scale = out_scale

        hidden_layers = kwargs.get("hidden_layers")
        self.dist = Normal
        self.mu = FcNet(in_features, out_features, hidden_layers=hidden_layers, **kwargs)
        self.log_std = FcNet(in_features, out_features, hidden_layers=hidden_layers, **kwargs)

        self.log_std_min = -10
        self.log_std_max = 2

        self._last_dist: Optional[Distribution] = None
        self._last_samples = None

    def log_prob(self, samples) -> Optional[torch.Tensor]:
        if self._last_dist is None:
            return None
        return self._last_dist.log_prob(samples).sum(axis=-1)

    def forward(self, x, deterministic: bool = False) -> torch.Tensor:
        mu = self.mu(x)
        if deterministic:
            self._last_dist, self._last_samples = (None, None)
            return mu

        log_std = self.log_std(x)
        log_std = torch.clamp(log_std, self.log_std_min, self.log_std_max)
        std = torch.exp(log_std)

        self._last_dist = dist = self.dist(mu, std)
        actions = dist.rsample()
        self._last_samples = {"mu": mu.squeeze(0), "std": std}
        return actions


class BetaPolicy(PolicyType):
    """
    Multivarate generalized version of the Dirichlet (1D) distribution.

    Uses torch.distributions.Beta or torch.distributions.Dirichlet
    distirubitions depending on the input size.

    https://pytorch.org/docs/stable/distributions.html#beta
    https://pytorch.org/docs/stable/distributions.html#dirichlet
    """

    param_dim = 2

    def __init__(self, size: int, bounds: Tuple[float, float] = (1, float("inf"))):
        """
        Parameters:
            size: Observation's dimensionality upon sampling.
            bounds: Beta dist input clamp for both alpha and betas.
                Both concentration are expected to be larger than 1.

        """
        super(BetaPolicy, self).__init__()
        self.bounds = bounds
        self.action_size = size
        self.dist = Beta if size > 1 else Dirichlet

    def forward(self, x) -> Distribution:
        x = x.view(-1, self.action_size, self.param_dim)
        x = torch.clamp(x, self.bounds[0], self.bounds[1])
        dist = self.dist(x[..., 0], x[..., 1])
        return dist

    @staticmethod
    def log_prob(dist, samples):
        return dist.log_prob(samples).mean(dim=-1)


class DirichletPolicy(PolicyType):

    param_dim = 1

    def __init__(self, *, alpha_min: float = 0.05):
        super(DirichletPolicy, self).__init__()
        self.alpha_min = alpha_min

    def forward(self, x) -> Distribution:
        x = torch.clamp(x, self.alpha_min)
        return Dirichlet(x)

    def log_prob(self, dist: Dirichlet, samples) -> torch.Tensor:
        return dist.log_prob(samples)


class DeterministicPolicy(PolicyType):

    param_dim = 1

    def __init__(self, action_size):
        super(DeterministicPolicy, self).__init__()
        self.action_size = action_size

    def forward(self, x):
        return x



import json
import logging
import os
import sys
from collections import deque
from pathlib import Path
from typing import List, Optional, Tuple

import numpy as np
import torch.multiprocessing as mp

from ai_traineree.agents import AgentBase
from ai_traineree.loggers import DataLogger
from ai_traineree.types import TaskType
from ai_traineree.types.experience import Experience

logging.basicConfig(
    stream=sys.stdout,
    level=logging.DEBUG,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

CMD_RESET = "RESET"
CMD_STOP = "STOP"


class MultiSyncEnvRunner:
    """Execute multiple environments/tasks concurrently with sync steps.

    All environments are distributed to separate processes. The MultiSyncEnvRunner
    acts as a manager that sends data between processes.

    Currently this class only supports training one agent at a time. The agent
    is expected handle stepping multiple steps at a time.
    """

    logger = logging.getLogger("MultiSyncEnvRunner")

    def __init__(self, tasks: List[TaskType], agent: AgentBase, max_iterations: int = int(1e5), **kwargs):
        """
        Expects the environment to come as the TaskType and the agent as the AgentBase.

        Keyword Arguments:
            window_len (int): Length of the score averaging window.
            data_logger: An instance of Data Logger, e.g. TensorboardLogger.
        """
        self.tasks = tasks
        self.task_num = len(tasks)
        self.num_processes = int(kwargs.get("processes", len(tasks)))
        self.processes = []
        self.parent_conns = []
        self.child_conns = []

        self.agent = agent
        self.max_iterations = max_iterations
        self.model_path = f"{tasks[0].name}_{agent.model}"
        self.state_dir = "run_states"
        self.logger.setLevel(kwargs.get("logger_level", logging.INFO))

        self.episode = 0
        self.iteration = 0
        self.all_scores = []
        self.all_iterations = []
        self.window_len = kwargs.get("window_len", 100)
        self.scores_window = deque(maxlen=self.window_len)

        self.data_logger: Optional[DataLogger] = kwargs.get("data_logger")
        self.logger.info("DataLogger: %s", str(self.data_logger))

    def __str__(self) -> str:
        return f"MultiSyncEnvRunner<{[t.name for t in self.tasks]}, {self.agent.model}>"

    def __del__(self):
        try:
            self.close_all()
        except Exception:
            self.logger.exception("Exception while clossing all connections")

    def reset(self):
        """Resets the EnvRunner. The task env and the agent are preserved."""
        self.episode = 0
        self.all_scores = []
        self.all_iterations = []
        self.scores_window = deque(maxlen=self.window_len)

    @staticmethod
    def step_task(conn, task):
        iteration = 0
        task.reset()

        # Infinite loop should be Ok because at the beginning of the loop
        # we wait for new data and at the end we send back.
        while True:
            received = conn.recv()

            if received == CMD_STOP:
                conn.close()
                return

            if received == CMD_RESET:
                conn.send(task.reset())
                iteration = 0
                continue

            t_idx, state, action = received
            iteration += 1
            task_out = task.step(action)
            next_state, reward, done, _ = task_out

            conn.send(
                {
                    "idx": t_idx,
                    "state": state,
                    "action": action,
                    "next_state": next_state,
                    "reward": reward,
                    "done": done,
                    "iteration": iteration,
                }
            )

    def init_processes(self) -> None:
        "Initiate agents in processes and establish communication"
        for p_idx in range(self.num_processes):
            parent_conn, child_conn = mp.Pipe()
            self.parent_conns.append(parent_conn)
            self.child_conns.append(child_conn)

            process = mp.Process(target=self.step_task, args=(child_conn, self.tasks[p_idx]))
            self.processes.append(process)

    def run(
        self,
        reward_goal: float = 100.0,
        max_episodes: int = 2000,
        max_iterations: int = int(1e6),
        eps_start: float = 1.0,
        eps_end: float = 0.01,
        eps_decay: float = 0.995,
        log_episode_freq: int = 1,
        checkpoint_every: Optional[int] = 200,
        force_new=False,
    ):
        """
        Evaluates the agent in the environment.
        The evaluation will stop when the agent reaches the `reward_goal` in the averaged last `self.window_len`, or
        when the number of episodes reaches the `max_episodes`.

        To help debugging one can set the `gif_every_episodes` to a positive integer which relates to how often a gif
        of the episode evaluation is written to the disk.

        Parameters:
            reward_goal: Goal to achieve on the average reward.
            max_episode: After how many episodes to stop regardless of the score.
            eps_start: Epsilon-greedy starting value.
            eps_end: Epislon-greeedy lowest value.
            eps_decay: Epislon-greedy decay value, eps[i+1] = eps[i] * eps_decay.
            log_episode_freq: Number of episodes between state logging.
            checkpoint_every: Number of episodes between storing the whole state, so that
                in case of failure it can be safely resumed from it.
            force_new: Flag whether to resume from previously stored state (False), or to
                start learning from a clean state (True).

        Returns:
            All obtained scores from all episodes.

        """

        # This method is mainly a wrapper around self._run to make it safer.
        # Somehow better option might be to add decorators but given unsure existance
        # of this class we'll refrain from doing so right now.
        try:
            # Initiate all processes and connections
            self.init_processes()

            return self._run(
                reward_goal=reward_goal,
                max_episodes=max_episodes,
                max_iterations=max_iterations,
                eps_start=eps_start,
                eps_end=eps_end,
                eps_decay=eps_decay,
                log_episode_freq=log_episode_freq,
                checkpoint_every=checkpoint_every,
                force_new=force_new,
            )

        finally:
            # All connections and processes need to be closed regardless of any exception.
            # Don't let get away. Who knows what zombie process are capable of?!
            self.close_all()

    def _run(
        self,
        reward_goal: float = 100.0,
        max_episodes: int = 2000,
        max_iterations: int = 1000,
        eps_start: float = 1.0,
        eps_end: float = 0.01,
        eps_decay: float = 0.995,
        log_episode_freq: int = 1,
        checkpoint_every: Optional[int] = 200,
        force_new: bool = False,
    ):
        # Initiate variables
        self.reset()
        self.epsilon = eps_start
        if not force_new:
            self.load_state(file_prefix=self.model_path)

        _scores = np.zeros(self.task_num)
        _iterations = np.zeros(self.task_num)

        obs = np.empty((len(self.tasks),) + self.tasks[0].obs_space.shape, dtype=np.float32)
        next_obs = obs.copy()
        actions = np.empty((len(self.tasks),) + self.tasks[0].action_space.shape, dtype=np.float32)

        # Start processes just before using them
        for process in self.processes:
            process.start()

        for idx, conn in enumerate(self.parent_conns):
            conn.send(CMD_RESET)
            reset_state = conn.recv()
            obs[idx] = reset_state

        mean_scores = []
        epsilons = []
        mean_score = -float("inf")

        while self.episode < max_episodes:
            self.iteration += 1

            experience = Experience(obs=obs)
            experience = self.agent.act(experience, self.epsilon)
            # TODO: Logic seems Ok.
            # This is likely the problem! Probably agent doesn't understand how to process multi agent data
            actions = experience.action if self.task_num != 1 else [experience.action]
            assert isinstance(actions, list), "For many agents needs to be list"
            experience.update(action=actions)

            # Step all tasks
            self._step_all_tasks(obs, actions)

            # Collect new observations from all stepped tasks
            new_experience, iterations = self._collect_all_tasks()
            experience = experience + new_experience
            next_obs = experience.next_obs
            _scores += experience.reward
            _iterations = iterations

            # All recently evaluated SARS are passed at the same time
            self.agent.step(experience)

            # Training part
            for idx in range(self.task_num):

                # Update Episode number if any agent is DONE or enough ITERATIONS
                if not (experience.done[idx] or _iterations[idx] >= max_iterations):
                    continue

                self.parent_conns[idx].send(CMD_RESET)
                next_obs[idx] = self.parent_conns[idx].recv()

                self.scores_window.append(_scores[idx])
                self.all_scores.append(_scores[idx])
                _scores[idx] = 0

                self.all_iterations.append(_iterations[idx])

                self.episode += 1
                mean_score: float = sum(self.scores_window) / len(self.scores_window)
                mean_scores.append(mean_score)
                epsilons.append(self.epsilon)

                # Log only once per evaluation, and outside s
                if self.episode % log_episode_freq == 0:
                    last_episodes = [self.episode - i for i in range(log_episode_freq)[::-1]]
                    self.info(
                        episodes=last_episodes,
                        iterations=self.all_iterations[-log_episode_freq:],
                        scores=self.all_scores[-log_episode_freq:],
                        mean_scores=mean_scores[-log_episode_freq:],
                        epsilons=epsilons[-log_episode_freq:],
                        loss=self.agent.loss,
                    )

                if checkpoint_every is not None and self.episode % checkpoint_every == 0:
                    self.save_state(self.model_path)

            obs = next_obs

            self.epsilon = max(eps_end, eps_decay * self.epsilon)

            if mean_score >= reward_goal and len(self.scores_window) == self.window_len:
                print(f"Environment solved after {self.episode} episodes!\tAverage Score: {mean_score:.2f}")
                self.save_state(self.model_path)
                self.agent.save_state(f"{self.model_path}_agent.net")
                break
        return self.all_scores

    def _step_all_tasks(self, obs, actions):
        for t_idx in range(self.task_num):
            self.parent_conns[t_idx].send((t_idx, obs[t_idx], actions[t_idx]))

    def _collect_all_tasks(self) -> Tuple[Experience, np.ndarray]:
        obs = np.empty((len(self.tasks),) + self.tasks[0].obs_space.shape, dtype=np.float32)
        next_obs = obs.copy()
        actions = np.empty((len(self.tasks),) + self.tasks[0].action_space.shape, dtype=np.float32)
        dones = np.empty(len(self.tasks))
        rewards = np.empty(len(self.tasks))
        iterations = np.empty(len(self.tasks))

        for t_idx in range(self.task_num):
            obj = self.parent_conns[t_idx].recv()

            idx = obj["idx"]
            rewards[idx] = obj["reward"]
            obs[idx] = obj["state"]
            actions[idx] = obj["action"]
            next_obs[idx] = obj["next_state"]
            dones[idx] = obj["done"]

            iterations[idx] = obj["iteration"]

        experience = Experience(obs=obs, next_obs=next_obs, action=actions, done=dones, reward=rewards)
        return experience, iterations

    def close_all(self):
        while self.parent_conns:
            conn = self.parent_conns.pop(0)
            conn.send(CMD_STOP)
            conn.close()

        while self.child_conns:
            self.child_conns.pop(0).close()

        while self.processes:
            p = self.processes.pop(0)
            p.terminate()
            p.join()

    def info(self, **kwargs):
        """
        Writes out current state into provided loggers.
        Writting to stdout is done through Python's logger, whereas all metrics
        are supposed to be handled via DataLogger. Currently supported are Tensorboard
        and Neptune (neptune.ai). To use one of these `data_logger` is expected.
        """
        if self.data_logger is not None:
            self.log_episode_metrics(**kwargs)
            self.log_data_interaction(**kwargs)
        if self.logger is not None:
            self.log_logger(**kwargs)

    def log_logger(self, **kwargs):
        """Writes out env logs via logger (either stdout or a file)."""
        if kwargs is None:
            return

        line_chunks = []
        episodes = kwargs.get("episodes")
        if episodes is not None:
            is_train = kwargs.get("train_mode", True)
            test_mark = "" if is_train else "T"
            line_chunks += [f"Episode {episodes[-1]}{test_mark}"]

        scores = kwargs.get("scores")
        if scores is not None:
            line_chunks += [f"Current Score: {scores[-1]:.2f}"]

        iterations = kwargs.get("iterations")
        if iterations is not None:
            line_chunks += [f"Iter: {iterations[-1]}"]

        mean_scores = kwargs.get("mean_scores")
        if mean_scores is not None and len(mean_scores):
            line_chunks += [f"Average Score: {mean_scores[-1]:.2f}"]

        epsilons = kwargs.get("epsilons")
        if epsilons is not None:
            line_chunks += [f"Epsilon: {epsilons[-1]:5.3f}"]

        loss = kwargs.get("loss", {})
        if loss is not None:
            line_chunks += [f"{loss_name.capitalize()}: {loss_value:10.4f}" for (loss_name, loss_value) in loss.items()]

        line = ";\t".join(line_chunks)
        self.logger.info(line)

    def log_episode_metrics(self, **kwargs):
        """Uses data_logger, e.g. Tensorboard, to store env metrics."""
        episodes: List[int] = kwargs.get("episodes", [])
        for episode, epsilon in zip(episodes, kwargs.get("epsilons", [])):
            self.data_logger.log_value("episode/epsilon", epsilon, episode)

        for episode, mean_score in zip(episodes, kwargs.get("mean_scores", [])):
            self.data_logger.log_value("episode/avg_score", mean_score, episode)

        for episode, score in zip(episodes, kwargs.get("scores", [])):
            self.data_logger.log_value("episode/score", score, episode)

        for episode, iteration in zip(episodes, kwargs.get("iterations", [])):
            self.data_logger.log_value("episode/iterations", iteration, episode)

    def log_data_interaction(self, **kwargs):
        if self.data_logger and hasattr(self.agent, "log_metrics"):
            self.agent.log_metrics(self.data_logger, self.iteration, full_log=kwargs.get("full_log", False))

    def save_state(self, state_name: str):
        """Saves the current state of the runner and the agent.

        Files are stored with appended episode number.
        Agents are saved with their internal saving mechanism.
        """
        state = {
            "tot_iterations": sum(self.all_iterations),
            "episode": self.episode,
            "epsilon": self.epsilon,
            "score": self.all_scores[-1],
            "average_score": sum(self.scores_window) / len(self.scores_window),
            "loss": self.agent.loss,
        }

        Path(self.state_dir).mkdir(parents=True, exist_ok=True)
        self.agent.save_state(f"{self.state_dir}/{state_name}_e{self.episode}.agent")
        with open(f"{self.state_dir}/{state_name}_e{self.episode}.json", "w") as f:
            json.dump(state, f)

    def load_state(self, file_prefix: str):
        """
        Loads state with the highest episode value for given agent and environment.
        """
        try:
            state_files = list(
                filter(lambda f: f.startswith(file_prefix) and f.endswith("json"), os.listdir(self.state_dir))
            )
            recent_episode_num = max([int(f[f.index("_e") + 2 : f.index(".")]) for f in state_files])
            state_name = [n for n in state_files if n.endswith(f"_e{recent_episode_num}.json")][0][:-5]
        except Exception:
            self.logger.warning("Couldn't load state. Forcing restart.")
            return

        self.logger.info("Loading saved state under: %s/%s.json", self.state_dir, state_name)
        with open(f"{self.state_dir}/{state_name}.json", "r") as f:
            state = json.load(f)
        self.episode = state.get("episode")
        self.epsilon = state.get("epsilon")

        self.all_scores.append(state.get("score"))
        self.all_iterations = []

        avg_score = state.get("average_score")
        for _ in range(min(self.window_len, self.episode)):
            self.scores_window.append(avg_score)

        self.logger.info("Loading saved agent state: %s/%s.agent", self.state_dir, state_name)
        self.agent.load_state(f"{self.state_dir}/{state_name}.agent")
        self.agent.loss = state.get("loss", 0)



import logging
from typing import Dict

from .data_logger import DataLogger

try:
    import neptune
except ImportError:
    msg = (
        "Couldn't import `neptune` module. This likely means that "
        "the `neptune` isn't installed, or that you're using a different Python environment. "
        "Installing `neptune` can be done for example by executing `pip install neptune`."
    )
    logging.exception(msg)
    raise ImportError(msg)


class NeptuneLogger(DataLogger):
    """Neptune.ai logger.

    Wrapper around the Neptune.ai logger.
    """

    name = "NeptuneLogger"

    def __init__(self, project_name: str, **kwargs):
        params = kwargs.pop("params", None)
        self.project = neptune.init(project_name, **kwargs)
        self.experiment = neptune.create_experiment(params=params)

    def __str__(self) -> str:
        return self.name

    def close(self):
        self.experiment.stop()

    def set_hparams(self, *, hparams: Dict):
        ...

    def log_value(self, name: str, value, step: int) -> None:
        self.experiment.log_metric(name, x=step, y=value)

    def log_values_dict(self, name: str, values, step: int) -> None:
        for _name, _value in values.items():
            self.log_value(f"{name}/{_name}", _value, step)

    def add_histogram(self, *args, **kwargs) -> None:
        return

    def create_histogram(self, name, values, step) -> None:
        return



import numpy as np

from ai_traineree.agents import AgentBase
from ai_traineree.types import DataSpace
from ai_traineree.types.experience import Experience


class DummyAgent(AgentBase):
    """Dummy Agent.

    Agent that returns random values in specified shapes.
    """

    model = "Dummy"

    def __init__(self, obs_space: DataSpace, action_space: DataSpace, **kwargs):
        """Initiates Dummy agent.

        Parameters:
            obs_space (DataSpace): Dataspace describing the input.
            action_space (DataSpace): Dataspace describing the output.

        """
        super().__init__(**kwargs)
        self.obs_space = obs_space
        self.action_space = action_space

        self.loss = {"loss": 0}

    @staticmethod
    def step(*args, **kwargs) -> None:
        """Simulates transition to the next internal stage."""
        pass

    def act(self, experience: Experience, *args, **kwargs) -> Experience:
        """Returns random discrete action and updates experience.

        Parameters:
            experience (Experience): Structure holding current iteration information.

        Returns:
            Experience updated in current action.

        """
        assert self.action_space.high is not None, "ActionSpace needs to have high boundary"

        action = np.random.randint(
            low=self.action_space.low, high=self.action_space.high + 1, size=self.action_space.shape
        )
        return experience.update(action=action)

    @staticmethod
    def log_metrics(*args, **kwargs) -> None:
        """Dummy method. Doesn't do anything."""
        pass

    @staticmethod
    def get_state(*args, **kwargs) -> None:
        """Dummy method. Doesn't do anything."""
        pass

    @staticmethod
    def save_state(*args, **kwargs) -> None:
        """Dummy method. Doesn't do anything."""
        pass

    @staticmethod
    def load_state(*args, **kwargs) -> None:
        """Dummy method. Doesn't do anything."""
        pass



# pylint: skip-file
# flake8: noqa
from .agent import AgentType, MultiAgentType
from .dataspace import DataSpace
from .experience import Experience
from .primitive import ActionType, DoneType, FeatureType, HyperparameterType, ObsType, RewardType, StateType
from .state import AgentState, BufferState, NetworkState
from .task import MultiAgentTaskType, TaskStepType, TaskType



from functools import cached_property
from typing import Dict

import torch
import torch.nn as nn
from torch.nn.functional import mse_loss
from torch.optim import Adam

from ai_traineree import DEVICE
from ai_traineree.agents import AgentBase
from ai_traineree.agents.agent_utils import hard_update, soft_update
from ai_traineree.buffers.replay import ReplayBuffer
from ai_traineree.loggers import DataLogger
from ai_traineree.networks.bodies import ActorBody, CriticBody
from ai_traineree.networks.heads import DoubleCritic
from ai_traineree.noise import GaussianNoise
from ai_traineree.types.dataspace import DataSpace
from ai_traineree.types.experience import Experience
from ai_traineree.utils import to_numbers_seq, to_tensor


class TD3Agent(AgentBase):
    """
    Twin Delayed Deep Deterministic (TD3) Policy Gradient.

    In short, it's a slightly modified/improved version of the DDPG. Compared to the DDPG in this package,
    which uses Guassian noise, this TD3 uses Ornstein–Uhlenbeck process as the noise.
    """

    model = "TD3"

    def __init__(
        self,
        obs_space: DataSpace,
        action_space: DataSpace,
        noise_scale: float = 0.5,
        noise_sigma: float = 1.0,
        **kwargs
    ):
        """
        Parameters:
            obs_space (DataSpace): Dataspace describing the input.
            action_space (DataSpace): Dataspace describing the output.
            noise_scale (float): Added noise amplitude. Default: 0.5.
            noise_sigma (float): Added noise variance. Default: 1.0.

        Keyword arguments:
            hidden_layers (tuple of ints): Shape of the hidden layers in fully connected network. Default: (128, 128).
            actor_lr (float): Learning rate for the actor (policy). Default: 0.003.
            critic_lr (float): Learning rate for the critic (value function). Default: 0.003.
            gamma (float): Discount value. Default: 0.99.
            tau (float): Soft-copy factor. Default: 0.02.
            actor_hidden_layers (tuple of ints): Shape of network for actor. Default: `hideen_layers`.
            critic_hidden_layers (tuple of ints): Shape of network for critic. Default: `hideen_layers`.
            max_grad_norm_actor (float) Maximum norm value for actor gradient. Default: 100.
            max_grad_norm_critic (float): Maximum norm value for critic gradient. Default: 100.
            batch_size (int): Number of samples used in learning. Default: 64.
            buffer_size (int): Maximum number of samples to store. Default: 1e6.
            warm_up (int): Number of samples to observe before starting any learning step. Default: 0.
            update_freq (int): Number of steps between each value function (critic) update. Default: 1.
            update_policy_freq (int): Number of steps between each policy (actor) update: Default: 2.
            number_updates (int): How many times to use learning step in the learning phase. Default: 1.

        """
        super().__init__(**kwargs)
        self.device = self._register_param(kwargs, "device", DEVICE)  # Default device is CUDA if available

        # Reason sequence initiation.
        assert len(action_space.shape) == 1, "Only 1D actions are supported"
        self.obs_space = obs_space
        self.action_space = action_space
        self._config["obs_space"] = self.obs_space
        self._config["action_space"] = self.action_space
        action_size = action_space.shape[0]

        hidden_layers = to_numbers_seq(self._register_param(kwargs, "hidden_layers", (128, 128)))
        self.actor = ActorBody(
            obs_space.shape,
            action_space.shape,
            hidden_layers=hidden_layers,
            gate=nn.ReLU(),
            gate_out=torch.tanh,
            device=self.device,
        )
        self.target_actor = ActorBody(
            obs_space.shape,
            action_space.shape,
            hidden_layers=hidden_layers,
            gate=nn.ReLU(),
            gate_out=torch.tanh,
            device=self.device,
        )

        self.critic = DoubleCritic(
            obs_space.shape, action_size, CriticBody, hidden_layers=hidden_layers, gate=nn.ReLU()
        ).to(self.device)
        self.target_critic = DoubleCritic(
            obs_space.shape, action_size, CriticBody, hidden_layers=hidden_layers, gate=nn.ReLU()
        ).to(self.device)

        # Noise sequence initiation
        self.noise = GaussianNoise(
            shape=action_space.to_feature(), mu=1e-8, sigma=noise_sigma, scale=noise_scale, device=self.device
        )

        # Optimization sequence initiation.
        self.actor_lr = float(self._register_param(kwargs, "actor_lr", 3e-4))
        self.critic_lr = float(self._register_param(kwargs, "critic_lr", 3e-4))
        self.actor_optimizer = Adam(self.actor.parameters(), lr=self.actor_lr)
        self.critic_optimizer = Adam(self.critic.parameters(), lr=self.critic_lr)
        self.max_grad_norm_actor: float = float(kwargs.get("max_grad_norm_actor", 10.0))
        self.max_grad_norm_critic: float = float(kwargs.get("max_grad_norm_critic", 10.0))

        self.gamma = float(self._register_param(kwargs, "gamma", 0.99))
        self.tau = float(self._register_param(kwargs, "tau", 0.02))
        self.batch_size = int(self._register_param(kwargs, "batch_size", 64))
        self.buffer_size = int(self._register_param(kwargs, "buffer_size", int(1e6)))
        self.buffer = ReplayBuffer(self.batch_size, self.buffer_size)

        self.warm_up = int(self._register_param(kwargs, "warm_up", 0))
        self.update_freq = int(self._register_param(kwargs, "update_freq", 1))
        self.update_policy_freq = int(self._register_param(kwargs, "update_policy_freq", 2))
        self.number_updates = int(self._register_param(kwargs, "number_updates", 1))

        # Breath, my child.
        self.reset_agent()
        self.iteration = 0
        self._loss_actor = float("nan")
        self._loss_critic = float("nan")

    @property
    def loss(self) -> Dict[str, float]:
        return {"actor": self._loss_actor, "critic": self._loss_critic}

    @loss.setter
    def loss(self, value):
        if isinstance(value, dict):
            self._loss_actor = value["actor"]
            self._loss_critic = value["critic"]
        else:
            self._loss_actor = value
            self._loss_critic = value

    def reset_agent(self) -> None:
        self.actor.reset_parameters()
        self.critic.reset_parameters()

        # Target sequence initiation
        hard_update(self.target_actor, self.actor)
        hard_update(self.target_critic, self.critic)

    @cached_property
    def action_min(self):
        return to_tensor(self.action_space.low)

    @cached_property
    def action_max(self):
        return to_tensor(self.action_space.high)

    @torch.no_grad()
    def act(self, experience: Experience, epsilon: float = 0.0) -> Experience:
        """
        Agent acting on observations.

        When the training_mode is True (default) a noise is added to each action.
        """
        # # Epsilon greedy
        # if self._rng.random() < epsilon:
        #     rnd = torch.rand(self.action_space.shape)
        #     rnd_actions = rnd * (self.action_max - self.action_min) - self.action_min
        #     action = rnd_actions.tolist()
        #     return experience.update(action=action)

        t_obs = to_tensor(experience.obs).float().to(self.device)
        action = self.actor(t_obs)
        if self.train:
            # action += self.noise.sample()
            noise = epsilon
            added_noise = noise * self.noise.sample()
            action += added_noise
            experience.update(noise=added_noise)
        action = torch.clamp(action, self.action_min, self.action_max).tolist()
        return experience.update(action=action)

    @property
    def _is_update_value(self):
        return self.iteration % self.update_freq == 0

    @property
    def _is_update_policy(self):
        return self.iteration % self.update_policy_freq == 0

    def step(self, experience: Experience) -> None:
        if not self.train:
            return

        self.iteration += 1
        self.buffer.add(
            obs=experience.obs,
            action=experience.action,
            reward=experience.reward,
            next_obs=experience.next_obs,
            done=experience.done,
        )

        if self.iteration < self.warm_up:
            return

        if len(self.buffer) <= self.batch_size:
            return

        if self._is_update_value or self._is_update_policy:
            for _ in range(self.number_updates):
                # Note: Inside this there's a delayed policy update.
                #       Every `update_policy_freq` it will learn `number_updates` times.
                self.learn(self.buffer.sample())

    def learn(self, experiences):
        """Update critics and actors"""
        rewards = to_tensor(experiences["reward"]).float().to(self.device).unsqueeze(1)
        dones = to_tensor(experiences["done"]).type(torch.int).to(self.device).unsqueeze(1)
        obss = to_tensor(experiences["obs"]).float().to(self.device)
        actions = to_tensor(experiences["action"]).to(self.device)
        next_obss = to_tensor(experiences["next_obs"]).float().to(self.device)

        if self._is_update_value:
            self._update_value_function(obss, actions, rewards, next_obss, dones)
            soft_update(self.target_actor, self.actor, self.tau)

        if self._is_update_policy:
            self._update_policy(obss)
            soft_update(self.target_critic, self.critic, self.tau)

    def _update_value_function(self, states, actions, rewards, next_states, dones):
        # Minimize the loss
        self.critic_optimizer.zero_grad()

        # critic loss
        next_actions = self.target_actor.act(next_states)
        with torch.no_grad():
            Q_target_next = torch.min(*self.target_critic.act(next_states, next_actions))
            Q_target = rewards + (self.gamma * Q_target_next * (1 - dones))
        Q1_expected, Q2_expected = self.critic(states, actions)
        loss_critic = mse_loss(Q1_expected, Q_target) + mse_loss(Q2_expected, Q_target)

        loss_critic.backward()
        nn.utils.clip_grad_norm_(self.critic.parameters(), self.max_grad_norm_critic)
        self.critic_optimizer.step()
        self._loss_critic = float(loss_critic.item())

    def _update_policy(self, states):
        self.critic.requires_grad_ = False
        self.actor_optimizer.zero_grad()

        # Compute actor loss
        pred_actions = self.actor(states)
        loss_actor = -self.critic(states, pred_actions)[0].mean()
        loss_actor.backward()
        nn.utils.clip_grad_norm_(self.actor.parameters(), self.max_grad_norm_actor)
        self.actor_optimizer.step()
        self._loss_actor = loss_actor.item()

        self.critic.requires_grad_ = True

    def state_dict(self) -> Dict[str, dict]:
        """Describes agent's networks.

        Returns:
            state: (dict) Provides actors and critics states.

        """
        return {
            "actor": self.actor.state_dict(),
            "target_actor": self.target_actor.state_dict(),
            "critic": self.critic.state_dict(),
            "target_critic": self.target_critic(),
        }

    def log_metrics(self, data_logger: DataLogger, step: int, full_log: bool = False):
        data_logger.log_value("loss/actor", self._loss_actor, step)
        data_logger.log_value("loss/critic", self._loss_critic, step)

    def get_state(self):
        return dict(
            actor=self.actor.state_dict(),
            target_actor=self.target_actor.state_dict(),
            critic=self.critic.state_dict(),
            target_critic=self.target_critic.state_dict(),
            config=self._config,
        )

    def save_state(self, path: str):
        agent_state = self.get_state()
        torch.save(agent_state, path)

    def load_state(self, path: str):
        agent_state = torch.load(path)
        self._config = agent_state.get("config", {})
        self.__dict__.update(**self._config)

        self.actor.load_state_dict(agent_state["actor"])
        self.critic.load_state_dict(agent_state["critic"])
        self.target_actor.load_state_dict(agent_state["target_actor"])
        self.target_critic.load_state_dict(agent_state["target_critic"])



import numbers

import mock

from ai_traineree.tasks import GymTask


def test_gym_task_actual_openai_discrete():
    # Assign
    gym_name = "CartPole-v1"

    # Act
    task = GymTask(gym_name)

    # Assert
    assert task.name == gym_name
    assert task.env is not None
    assert task.can_render is True
    assert task.is_discrete is True
    assert task.obs_size == 4
    assert task.action_size == 2


def test_gym_task_actual_openai_continious():
    # Assign
    gym_name = "Pendulum-v0"

    # Act
    task = GymTask(gym_name, can_render=False)

    # Assert
    assert task.name == gym_name
    assert task.env is not None
    assert task.can_render is False
    assert task.is_discrete is False
    assert task.obs_size == 3
    assert task.action_size == 1


@mock.patch("ai_traineree.tasks.gym")
def test_gym_task_reset(mock_gym, fix_env):
    # Assign
    mock_gym.make.return_value = fix_env
    task = GymTask("example")

    # Act
    out = task.reset()

    # Assert
    assert fix_env.reset.called_once()
    assert len(out) > 0


@mock.patch("ai_traineree.tasks.gym")
def test_gym_task_step(mock_gym, fix_env):
    # Assign
    mock_gym.make.return_value = fix_env
    task = GymTask("example")
    action = 2.13

    # Act
    out = task.step(action=action)

    # Assert
    assert fix_env.step.called_once_with(int(action))
    assert len(out) == 4
    assert hasattr(out[0], "__iter__")
    assert isinstance(out[1], numbers.Number)
    assert isinstance(out[2], bool)
    assert isinstance(out[3], str)


@mock.patch("ai_traineree.tasks.gym")
def test_gym_task_step_discrete(mock_gym, fix_env_discrete):
    # Assign
    mock_gym.make.return_value = fix_env_discrete
    task = GymTask("example")
    action = 2.0

    # Act
    out = task.step(action=action)

    # Assert
    assert fix_env_discrete.step.called_once_with(int(action))
    assert len(out) == 4
    assert hasattr(out[0], "__iter__")
    assert isinstance(out[1], numbers.Number)
    assert isinstance(out[2], bool)
    assert isinstance(out[3], str)


@mock.patch("ai_traineree.tasks.gym")
def test_gym_task_render(mock_gym, fix_env):
    # Assign
    mock_gym.make.return_value = fix_env
    task = GymTask("CanRender", can_render=True)

    # Act
    task.render()

    # Assert
    assert fix_env.render.called_once_with("rgb_array")


@mock.patch("ai_traineree.tasks.gym")
def test_gym_task_render_human(mock_gym, fix_env):
    # Assign
    mock_gym.make.return_value = fix_env
    task = GymTask("CanRender", can_render=True)

    # Act
    task.render(mode="human")

    # Assert
    assert fix_env.render.called_once_with("human")


@mock.patch("ai_traineree.tasks.gym")
def test_gym_task_render_cannot_render(mock_gym, fix_env):
    # Assign
    mock_gym.make.return_value = fix_env
    task = GymTask("CanRender", can_render=False)

    # Act
    task.render()

    # Assert
    assert not fix_env.render.called


def test_gym_seed():
    def _deterministic_gym_eval(task):
        states = []
        for _ in range(3):
            states.append(task.reset())
            for _ in range(3):
                states.append(task.step(0)[0])
        return states

    # Assign
    task = GymTask("CartPole-v1")

    # Act
    task.seed(0)
    first_states = _deterministic_gym_eval(task)
    task.seed(0)
    second_states = _deterministic_gym_eval(task)

    # Assert
    for state_1, state_2 in zip(first_states, second_states):
        assert all([s1 == s2 for (s1, s2) in zip(state_1, state_2)])



import numpy as np
import pylab as plt
import torch

from ai_traineree.agents.dqn import DQNAgent
from ai_traineree.loggers.file_logger import FileLogger
from ai_traineree.runners.env_runner import EnvRunner
from ai_traineree.tasks import GymTask

seed = 32167
# torch.set_deterministic(True)
torch.manual_seed(seed)
data_logger = FileLogger(filepath="cart_dqn")

env_name = "CartPole-v1"
task = GymTask(env_name, seed=seed)
agent = DQNAgent(task.obs_space, task.action_space, n_steps=5, seed=seed)
env_runner = EnvRunner(task, agent, data_logger=data_logger, seed=seed)

scores = env_runner.run(reward_goal=100, max_episodes=300, force_new=True)
env_runner.interact_episode(render=True)

# plot the scores
fig = plt.figure()
ax = fig.add_subplot(111)
plt.plot(np.arange(len(scores)), scores)
plt.ylabel("Score")
plt.xlabel("Episode #")
plt.savefig(f"{env_name}.png", dpi=120)
plt.show()



import json
import random

import numpy as np
import pytest
import torch

from ai_traineree.types.dataspace import DataSpace
from ai_traineree.utils import (
    condens_ndarray,
    serialize,
    str_to_list,
    str_to_number,
    str_to_seq,
    str_to_tuple,
    to_numbers_seq,
    to_tensor,
)
from conftest import deterministic_interactions


def generate_value():
    if random.random() < 0.5:
        return [random.random() for _ in range(random.randint(0, 10))]
    return random.random()


def test_to_tensor_tensor():
    # Assign
    t = torch.tensor([0, 1, 2, 3])

    # Act
    new_t = to_tensor(t)

    # Assert
    assert torch.equal(new_t, t)


def test_to_tensor_list():
    # Assign
    int_l = [0, 1, 2, 3]
    float_l = [0.5, 1.1, 2.9, 3.0]
    int_l_shape = [list(range(4 * i, 4 * (i + 1))) for i in range(5)]

    # Act
    int_t = to_tensor(int_l)
    float_t = to_tensor(float_l)
    int_t_shape = to_tensor(int_l_shape)

    # Assert
    assert torch.equal(torch.tensor(int_l), int_t)
    assert torch.equal(torch.tensor(float_l), float_t)
    assert torch.equal(torch.tensor(int_l_shape), int_t_shape)

    assert int_t.shape == (4,)
    assert float_t.shape == (4,)
    assert int_t_shape.shape == (5, 4)


def test_to_tensor_list_tesors():
    # Assign
    int_l = [torch.tensor([0, 1, 2, 3]), torch.tensor([1, 2, 5, 10])]
    float_l = [torch.tensor([0.5, 1.1, 2.9, 3.0]), torch.tensor([0.1, 0.1, 0.1, 10.0])]

    # Act
    int_t = to_tensor(int_l)
    float_t = to_tensor(float_l)

    # Assert
    assert torch.equal(torch.stack(int_l), int_t)
    assert torch.equal(torch.stack(float_l), float_t)

    assert int_t.shape == (2, 4)
    assert float_t.shape == (2, 4)


def test_to_tensor_numpy():
    # Assign
    int_l = np.array([0, 1, 2, 3])
    float_l = np.random.random(4)
    float_l = np.random.random(4)

    # Act
    int_t = to_tensor(int_l)
    float_t = to_tensor(float_l)

    # Assert
    assert torch.equal(torch.tensor(int_l), int_t)
    assert torch.equal(torch.tensor(float_l), float_t)


def test_str_to_number():
    assert 1 == str_to_number("1")
    assert 1 == str_to_number(" 1 ")
    assert 1.5 == str_to_number(" 1.5 ")

    with pytest.raises(ValueError):
        str_to_list("[0, 1")


def test_str_to_list():
    assert [0.1, 0, 0, 2.5, 1] == str_to_list("[0.1, 0, 0, 2.5, 1]")
    assert [0.1, 1] == str_to_list("[0.1, 1]")
    assert [0, 1] == str_to_list("[0,1]")

    with pytest.raises(ValueError):
        str_to_list("[0, 1")
    with pytest.raises(ValueError):
        str_to_list("0, 1]")
    with pytest.raises(ValueError):
        str_to_list("[0, 1)")


def test_str_to_tuple():
    assert (1, 4) == str_to_tuple("(1, 4)")
    assert (1, 4) == str_to_tuple("1, 4")
    assert (1, 4) == str_to_tuple("1,4      ")
    assert (4,) == str_to_tuple("4")

    with pytest.raises(ValueError):
        str_to_tuple("(0, 1")
    with pytest.raises(ValueError):
        str_to_tuple("0, 1)")
    with pytest.raises(ValueError):
        str_to_tuple("[0, 1)")


def test_str_to_seq():
    assert [0.1, 0] == str_to_seq("[0.1, 0]") == str_to_list("[0.1, 0]")
    assert [0, 1] == str_to_list("[0,1]") == str_to_list("[0,1]")
    assert (1, 4) == str_to_seq("(1, 4)") == str_to_tuple("(1, 4)")
    assert (1, 4) == str_to_seq("1, 4") == str_to_tuple("1, 4")
    assert (1, 4) == str_to_seq("1,4        ") == str_to_tuple("1,4      ")
    assert (4,) == str_to_seq("4")

    with pytest.raises(ValueError):
        str_to_seq("(0, 1")
    with pytest.raises(ValueError):
        str_to_seq("0, 1)")
    with pytest.raises(ValueError):
        str_to_seq("[0, 1)")
    with pytest.raises(ValueError):
        str_to_seq("{0, 1}")


def test_serialize_experience_individual():
    from ai_traineree.types.experience import Experience

    for key in Experience.whitelist:
        value = generate_value()
        e = Experience(**{key: value})

        ser = serialize(e)
        assert ser == '{"%s": "%s"}' % (key, value)


def test_serialize_experience_all_keys():
    import json

    from ai_traineree.types.experience import Experience

    state = {key: generate_value() for key in Experience.whitelist}
    e = Experience(**state)

    # Act
    ser = serialize(e)

    # Assert
    assert json.loads(ser) == {k: str(v) for (k, v) in state.items()}


def test_serialize_buffer_state_list():
    from ai_traineree.buffers import BufferState
    from ai_traineree.types.experience import Experience

    def generate_exp() -> Experience:
        return Experience(
            state=np.random.random(10).tolist(),
            next_state=np.random.random(10).tolist(),
            action=np.random.random(3).tolist(),
            reward=float(random.random()),
            done=bool(random.randint(0, 8) == 8),
        )

    # Assign
    buffer_size = 1e1
    buffer_state = BufferState(type="Type", batch_size=200, buffer_size=buffer_size)
    buffer_state.data = [generate_exp() for _ in range(int(buffer_size))]

    # Act
    ser = serialize(buffer_state)

    # Assert
    des = json.loads(ser)
    assert set(des.keys()) == set(("type", "buffer_size", "batch_size", "data", "extra"))
    assert len(des["data"]) == buffer_size
    assert des["type"] == "Type"
    assert des["extra"] is None


def test_serialize_buffer_state_numpy():
    from ai_traineree.buffers import BufferState
    from ai_traineree.types.experience import Experience

    def generate_exp() -> Experience:
        return Experience(
            state=np.random.random(10),
            next_state=np.random.random(10),
            action=np.random.random(3),
            reward=random.random(),
            done=(random.randint(0, 8) == 8),
        )

    # Assign
    buffer_size = int(1e1)
    buffer_state = BufferState(type="Type", batch_size=200, buffer_size=buffer_size)
    buffer_state.data = [generate_exp() for _ in range(int(buffer_size))]

    # Act
    ser = serialize(buffer_state)

    # Assert
    des = json.loads(ser)
    assert set(des.keys()) == set(("type", "buffer_size", "batch_size", "data", "extra"))
    assert len(des["data"]) == buffer_size
    assert des["type"] == "Type"
    assert des["extra"] is None


def test_serialize_network_state_actual():
    from ai_traineree.agents.dqn import DQNAgent

    agent = DQNAgent(DataSpace(dtype="float", shape=(4,)), DataSpace("int", (1,), low=0, high=2))
    deterministic_interactions(agent, 30)
    network_state = agent.get_network_state()

    # Act
    ser = serialize(network_state)

    # Assert
    des = json.loads(ser)
    assert set(des["net"].keys()) == set(("target_net", "net"))


def test_serialize_agent_state_actual():
    from ai_traineree.agents.dqn import DQNAgent

    agent = DQNAgent(DataSpace(dtype="float", shape=(4,)), DataSpace("int", (1,), low=0, high=2))
    deterministic_interactions(agent, 30)
    state = agent.get_state()

    # Act
    ser = serialize(state)

    # Assert
    des = json.loads(ser)
    assert des["model"] == DQNAgent.model
    assert len(des["buffer"]["data"]) == 30
    assert set(des["network"]["net"].keys()) == set(("target_net", "net"))


def test_to_numbers_seq_tuple_list():
    t_in = (1, 2, 3)
    l_in = [1, 2, 3]

    t_out = to_numbers_seq(t_in)
    l_out = to_numbers_seq(l_in)

    assert t_out == t_in
    assert l_out == l_in


def test_to_numbers_seq_str():
    s_t1_in = "(1,2,3)"
    s_t2_in = "(1, 2, 3)"
    s_l1_in = "[1,2,3]"
    s_l2_in = "[1, 2, 3]"

    s_t1_out = to_numbers_seq(s_t1_in)
    s_t2_out = to_numbers_seq(s_t2_in)
    s_l1_out = to_numbers_seq(s_l1_in)
    s_l2_out = to_numbers_seq(s_l2_in)

    assert s_t1_out == (1, 2, 3)
    assert s_t2_out == (1, 2, 3)
    assert s_l1_out == [1, 2, 3]
    assert s_l2_out == [1, 2, 3]


def test_to_numbers_seq_num():
    int_in = 2
    float_in = 2.123

    int_out = to_numbers_seq(int_in)
    float_out = to_numbers_seq(float_in)

    assert int_out == (2,)
    assert float_out == (2.123,)


def test_to_numbers_seq_unknown():
    test_cases = ["asd", "[2,1", {"key": "value"}]

    for test_case in test_cases:
        with pytest.raises(ValueError):
            to_numbers_seq(test_case)


def test_condens_ndarray_possible():
    l_shape = []
    for shape_idx in range(5, 1):
        l_shape.append(shape_idx)
        shape = tuple(l_shape)
        a = np.zeros(shape)
        assert 0 == condens_ndarray(a)

        a = np.ones(shape) * 1.5
        assert 0.5 == condens_ndarray(a)


def test_condens_ndarray_not_possible():
    a = np.arange(10)
    assert np.all(a.copy() == condens_ndarray(a.copy()))

    a = np.random.random((3, 4, 2))
    assert np.all(a.copy() == condens_ndarray(a.copy()))


if __name__ == "__main__":
    test_serialize_network_state_actual()



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Tuple, Union

import gym
from torch import Tensor

from ai_traineree.types.primitive import FeatureType, Numeric
from ai_traineree.utils import condens_ndarray, to_numbers_seq


@dataclass
class DataSpace:
    dtype: str
    shape: Tuple[int]
    low: Optional[Union[Numeric, Tensor]] = None
    high: Optional[Union[Numeric, Tensor]] = None

    @staticmethod
    def from_int(size: int):
        return DataSpace(dtype="int", shape=(size,), low=0, high=(size - 1))

    @staticmethod
    def from_gym_space(space) -> DataSpace:
        if isinstance(space, gym.spaces.multi_discrete.MultiDiscrete):
            return DataSpace(
                dtype=str(space.dtype),
                shape=(len(space.nvec),),
                low=0,
                high=to_numbers_seq(space.nvec),
            )
        elif "Discret" in str(space):
            return DataSpace(
                dtype=str(space.dtype),
                shape=(1,),
                low=0,
                high=space.n - 1,  # Inclusive bounds, so n=2 -> [0,1]
            )
        else:
            low = condens_ndarray(space.low)
            if not isinstance(low, (int, float)):
                low = low.tolist()
            high = condens_ndarray(space.high)
            if not isinstance(high, (int, float)):
                high = high.tolist()
            return DataSpace(
                shape=space.shape,
                dtype=str(space.dtype),
                low=low,
                high=high,
            )

    def to_feature(self) -> FeatureType:
        """Extracts required FeatureType from DataSpace.

        The reason this method exists is because DataSpace can be both discrete and continuous.
        Even though in both cases dataspace shape represent tensor shape of the space, it is usually used
        differently depending on the problem. Methods that work with discrete spaces will need to select
        on of the discrete values meaning that they need to know the total size (high - low),
        whereas in case of continuous space the shape is enough.

        Likely this method isn't necessary with some cleaver dataspace arrangement or additional property.
        However, for now, it simplifies.
        """
        if self.dtype.startswith("int") and len(self.shape) == 1:
            return (int(self.high) - int(self.low) + 1,)
        return self.shape



from collections import OrderedDict, defaultdict
from typing import Any, Dict, List, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

from ai_traineree import DEVICE
from ai_traineree.agents.agent_utils import hard_update, soft_update
from ai_traineree.agents.ddpg import DDPGAgent
from ai_traineree.buffers.replay import ReplayBuffer
from ai_traineree.loggers import DataLogger
from ai_traineree.networks.bodies import CriticBody
from ai_traineree.types import MultiAgentType
from ai_traineree.types.dataspace import DataSpace
from ai_traineree.types.experience import Experience
from ai_traineree.utils import to_numbers_seq, to_tensor


class MADDPGAgent(MultiAgentType):

    model = "MADDPG"

    def __init__(self, obs_space: DataSpace, action_space: DataSpace, num_agents: int, **kwargs):
        """Initiation of the Multi Agent DDPG.

        All keywords are also passed to DDPG agents.

        Parameters:
            obs_size (int): Dimensionality of the state.
            action_size (int): Dimensionality of the action.
            num_agents (int): Number of agents.

        Keyword Arguments:
            hidden_layers (tuple of ints): Shape for fully connected hidden layers.
            noise_scale (float): Default: 1.0. Noise amplitude.
            noise_sigma (float): Default: 0.5. Noise variance.
            actor_lr (float): Default: 0.001. Learning rate for actor network.
            critic_lr (float): Default: 0.001. Learning rate for critic network.
            gamma (float): Default: 0.99. Discount value
            tau (float): Default: 0.02. Soft copy value.
            gradient_clip (optional float): Max norm for learning gradient. If None then no clip.
            batch_size (int): Number of samples per learning.
            buffer_size (int): Number of previous samples to remember.
            warm_up (int): Number of samples to see before start learning.
            update_freq (int): How many samples between learning sessions.
            number_updates (int): How many learning cycles per learning session.

        """

        self.device = self._register_param(kwargs, "device", DEVICE, update=True)
        self.obs_space = obs_space
        self.action_space = action_space
        self.num_agents: int = num_agents
        self.agent_names: List[str] = kwargs.get("agent_names", map(str, range(self.num_agents)))

        hidden_layers = to_numbers_seq(self._register_param(kwargs, "hidden_layers", (100, 100), update=True))
        noise_scale = float(self._register_param(kwargs, "noise_scale", 0.5))
        noise_sigma = float(self._register_param(kwargs, "noise_sigma", 1.0))
        actor_lr = float(self._register_param(kwargs, "actor_lr", 3e-4))
        critic_lr = float(self._register_param(kwargs, "critic_lr", 3e-4))

        self.agents: Dict[str, DDPGAgent] = OrderedDict(
            {
                agent_name: DDPGAgent(
                    obs_space,
                    action_space,
                    actor_lr=actor_lr,
                    critic_lr=critic_lr,
                    noise_scale=noise_scale,
                    noise_sigma=noise_sigma,
                    **kwargs,
                )
                for agent_name in self.agent_names
            }
        )

        self.gamma = float(self._register_param(kwargs, "gamma", 0.99))
        self.tau = float(self._register_param(kwargs, "tau", 0.02))
        self.gradient_clip: Optional[float] = self._register_param(kwargs, "gradient_clip")

        self.batch_size = int(self._register_param(kwargs, "batch_size", 64))
        self.buffer_size = int(self._register_param(kwargs, "buffer_size", int(1e6)))
        self.buffer = ReplayBuffer(self.batch_size, self.buffer_size)

        self.warm_up = int(self._register_param(kwargs, "warm_up", 0))
        self.update_freq = int(self._register_param(kwargs, "update_freq", 1))
        self.number_updates = int(self._register_param(kwargs, "number_updates", 1))

        assert len(obs_space.shape) == 1, "Only 1D obs spaces are supported now"
        assert len(action_space.shape) == 1, "Only 1D action spaces are supported now"
        ma_obs_shape = (num_agents * obs_space.shape[0],)
        ma_action_size = num_agents * action_space.shape[0]
        self.critic = CriticBody(ma_obs_shape, ma_action_size, hidden_layers=hidden_layers).to(self.device)
        self.target_critic = CriticBody(ma_obs_shape, ma_action_size, hidden_layers=hidden_layers).to(self.device)
        self.critic_optimizer = optim.Adam(self.critic.parameters(), lr=critic_lr)
        hard_update(self.target_critic, self.critic)

        self._step_data = {}
        self._loss_critic: float = float("nan")
        self._loss_actor: Dict[str, float] = {name: float("nan") for name in self.agent_names}
        self.reset()

    @property
    def loss(self) -> Dict[str, float]:
        out = {}
        for agent_name, agent in self.agents.items():
            for loss_name, loss_value in agent.loss.items():
                out[f"{agent_name}_{loss_name}"] = loss_value
            out[f"{agent_name}_actor"] = self._loss_actor[agent_name]
        out["critic"] = self._loss_critic
        return out

    def reset(self):
        self.iteration = 0
        self.reset_agents()

    def reset_agents(self):
        for agent in self.agents.values():
            agent.reset_agent()
        self.critic.reset_parameters()
        self.target_critic.reset_parameters()

    def step(self, agent_name: str, experience: Experience) -> None:
        self._step_data[agent_name] = experience

    def commit(self):
        step_data = defaultdict(list)
        for agent in self.agents:
            agent_experience: Experience = self._step_data[agent]
            step_data["obs"].append(agent_experience.obs)
            step_data["action"].append(agent_experience.action)
            step_data["reward"].append(agent_experience.reward)
            step_data["next_obs"].append(agent_experience.next_obs)
            step_data["done"].append(agent_experience.done)

        self.buffer.add(**step_data)
        self._step_data = {}
        self.iteration += 1
        if self.iteration < self.warm_up:
            return

        if len(self.buffer) > self.batch_size and (self.iteration % self.update_freq) == 0:
            for _ in range(self.number_updates):
                samples = self.buffer.sample()
                for agent_name in self.agents:
                    self.learn(samples, agent_name)
            self.update_targets()

    @torch.no_grad()
    def act(self, agent_name: str, experience: Experience, noise: float = 0.0) -> Experience:
        """Get actions from all agents. Synchronized action.

        Parameters:
            states: List of states per agent. Positions need to be consistent.
            noise: Scale for the noise to include

        Returns:
            actions: List of actions that each agent wants to perform

        """
        experience.update(obs=experience.obs)
        agent = self.agents[agent_name]
        experience = agent.act(experience, noise)
        return experience

    def __flatten_actions(self, actions):
        return actions.view(-1, self.num_agents * self.action_space.shape[0])

    def learn(self, experiences, agent_name: str) -> None:
        """update the critics and actors of all the agents"""
        ma_obs_size = self.num_agents * self.obs_space.shape[0]
        ma_action_size = self.num_agents * self.action_space.shape[0]

        # TODO: Just look at this mess.
        agent_number = list(self.agents).index(agent_name)
        agent_rewards = to_tensor(experiences["reward"]).select(1, agent_number).unsqueeze(-1).float().to(self.device)
        agent_dones = (
            to_tensor(experiences["done"]).select(1, agent_number).unsqueeze(-1).type(torch.int).to(self.device)
        )

        obss = (
            to_tensor(experiences["obs"])
            .to(self.device)
            .view((self.batch_size, self.num_agents) + self.obs_space.shape)
        )
        actions = to_tensor(experiences["action"]).to(self.device)
        next_obss = (
            to_tensor(experiences["next_obs"])
            .float()
            .to(self.device)
            .view((self.batch_size, self.num_agents) + self.obs_space.shape)
        )

        flat_obss = obss.view(-1, ma_obs_size)
        flat_next_obss = next_obss.view(-1, ma_obs_size)
        flat_actions = actions.view(-1, ma_action_size)
        assert agent_rewards.shape == agent_dones.shape == (self.batch_size, 1)
        assert obss.shape == next_obss.shape == (self.batch_size, self.num_agents) + self.obs_space.shape
        assert actions.shape == (self.batch_size, self.num_agents) + self.action_space.shape
        assert flat_actions.shape == (self.batch_size, ma_action_size)

        agent = self.agents[agent_name]

        next_actions = actions.detach().clone()
        next_actions.data[:, agent_number] = agent.target_actor(next_obss[:, agent_number, :])
        assert next_actions.shape == (self.batch_size, self.num_agents) + self.action_space.shape

        # critic loss
        Q_target_next = self.target_critic(flat_next_obss, self.__flatten_actions(next_actions))
        Q_target = agent_rewards + (self.gamma * Q_target_next * (1 - agent_dones))
        Q_expected = self.critic(flat_obss, flat_actions)
        loss_critic = F.mse_loss(Q_expected, Q_target)

        # Minimize the loss
        self.critic_optimizer.zero_grad()
        loss_critic.backward()
        if self.gradient_clip:
            nn.utils.clip_grad_norm_(self.critic.parameters(), self.gradient_clip)
        self.critic_optimizer.step()
        self._loss_critic = float(loss_critic.mean().item())

        # Compute actor loss
        pred_actions = actions.detach().clone()
        # pred_actions.data[:, agent_number] = agent.actor(flat_obss)
        pred_actions.data[:, agent_number] = agent.actor(obss[:, agent_number, :])

        loss_actor = -self.critic(flat_obss, self.__flatten_actions(pred_actions)).mean()
        agent.actor_optimizer.zero_grad()
        loss_actor.backward()
        agent.actor_optimizer.step()
        self._loss_actor[agent_name] = loss_actor.mean().item()

    def update_targets(self):
        """soft update targets"""
        for agent in self.agents.values():
            soft_update(agent.target_actor, agent.actor, self.tau)
        soft_update(self.target_critic, self.critic, self.tau)

    def log_metrics(self, data_logger: DataLogger, step: int, full_log: bool = False):
        data_logger.log_value("loss/critic", self._loss_critic, step)
        for agent_name, agent in self.agents.items():
            data_logger.log_values_dict(f"{agent_name}/loss", agent.loss, step)

    def get_state(self) -> Dict[str, dict]:
        """Returns agents' internal states"""
        agents_state = {}
        agents_state["config"] = self._config
        for agent_name, agent in self.agents.items():
            agents_state[agent_name] = {"state": agent.state_dict(), "config": agent._config}
        return agents_state

    def save_state(self, path: str):
        """Saves current state of the Multi Agent instance and all related agents.

        All states are stored via PyTorch's :func:`save <torch.save>` function.

        Parameters:
            path: (str) String path to a location where the state is store.

        """
        agents_state = self.get_state()
        torch.save(agents_state, path)

    def load_state(self, *, path: Optional[str] = None, agent_state: Optional[dict] = None) -> None:
        """Loads the state into the Multi Agent.

        The state can be provided either via path to a file that contains the state,
        see :meth:`save_state <self.save_state>`, or direclty via `state`.

        Parameters:
            path: (str) A path where the state was saved via `save_state`.
            state: (dict) Already loaded state kept in memory.

        """
        if path is None and agent_state is None:
            raise ValueError("Either `path` or `agent_state` must be provided to load agent's state.")
        if path is not None and agent_state is None:
            agent_state = torch.load(path)
        assert agent_state is not None, "Can't load state if neither agent state or valid path is provided"

        self._config = agent_state.get("config", {})
        self.__dict__.update(**self._config)
        for agent_name, agent in self.agents.items():
            _agent_state = agent_state[agent_name]
            agent.load_state(agent_state=_agent_state["state"])
            agent._config = _agent_state["config"]
            agent.__dict__.update(**agent._config)

    def seed(self, seed: int) -> None:
        for agent in self.agents.values():
            agent.seed(seed)

    def state_dict(self) -> Dict[str, Any]:
        return {name: agent.state_dict() for (name, agent) in self.agents.items()}



"""
This 'Test suite' simply makes sure that all/most agents run.
Runs might not be perfect, might not be meaningful but at least
the code isn't broken.

**DO NOT TREAT THIS AS A UNIT TEST**
"""

from ai_traineree.agents.d3pg import D3PGAgent
from ai_traineree.agents.ddpg import DDPGAgent
from ai_traineree.agents.dqn import DQNAgent
from ai_traineree.agents.ppo import PPOAgent
from ai_traineree.agents.rainbow import RainbowAgent
from ai_traineree.agents.sac import SACAgent
from ai_traineree.agents.td3 import TD3Agent
from ai_traineree.runners.env_runner import EnvRunner
from ai_traineree.tasks import GymTask

DEVICE = "cpu"


# Discrete agents
def test_runs_dqn():
    # Assign
    task = GymTask("CartPole-v1")
    agent = DQNAgent(task.obs_space, task.action_space, device=DEVICE)
    env_runner = EnvRunner(task, agent, max_iterations=50)

    # Act
    env_runner.run(reward_goal=10, max_episodes=10, force_new=True)


def test_runs_rainbow():
    # Assign
    task = GymTask("CartPole-v1")
    agent = RainbowAgent(task.obs_space, task.action_space, device=DEVICE)
    env_runner = EnvRunner(task, agent, max_iterations=50)

    # Act
    env_runner.run(reward_goal=10, max_episodes=10, force_new=True)


def test_runs_ppo():
    # Assign
    task = GymTask("Pendulum-v0")
    agent = PPOAgent(task.obs_space, task.action_space, device=DEVICE)
    env_runner = EnvRunner(task, agent, max_iterations=50)

    # Act
    env_runner.run(reward_goal=10, max_episodes=10, force_new=True)


def test_runs_ddpg():
    # Assign
    task = GymTask("Pendulum-v0")
    agent = DDPGAgent(task.obs_space, task.action_space, device=DEVICE)
    env_runner = EnvRunner(task, agent, max_iterations=50)

    # Act
    env_runner.run(reward_goal=10, max_episodes=10, force_new=True)


def test_runs_td3():
    # Assign
    task = GymTask("Pendulum-v0")
    agent = TD3Agent(task.obs_space, task.action_space, device=DEVICE)
    env_runner = EnvRunner(task, agent, max_iterations=50)

    # Act
    env_runner.run(reward_goal=10, max_episodes=10, force_new=True)


def test_runs_sac():
    # Assign
    task = GymTask("BipedalWalker-v3")
    agent = SACAgent(task.obs_space, task.action_space, device=DEVICE)
    env_runner = EnvRunner(task, agent, max_iterations=50)

    # Act
    env_runner.run(reward_goal=10, max_episodes=10, force_new=True)


def test_runs_d3pg():
    # Assign
    task = GymTask("BipedalWalker-v3")
    agent = D3PGAgent(task.obs_space, task.action_space, device=DEVICE)
    env_runner = EnvRunner(task, agent, max_iterations=50)

    # Act
    env_runner.run(reward_goal=10, max_episodes=10, force_new=True)



import abc
from typing import Any, List, Optional, Tuple

from .dataspace import DataSpace
from .primitive import ActionType, DoneType, RewardType, StateType

TaskStepType = Tuple[StateType, RewardType, DoneType, Any]


class TaskType(abc.ABC):
    """
    .. _TaskType

    """

    name: str
    action_space: DataSpace
    obs_space: DataSpace
    is_discrete: bool

    @abc.abstractmethod
    def seed(self, seed):
        pass

    @abc.abstractmethod
    def step(self, action: ActionType, **kwargs) -> TaskStepType:
        pass

    @abc.abstractmethod
    def render(self, mode: Optional[str] = None) -> None:
        pass

    @abc.abstractmethod
    def reset(self) -> StateType:
        pass


class MultiAgentTaskType(TaskType):
    @abc.abstractmethod
    def reset(self) -> List[StateType]:
        pass



import copy
from typing import Callable, Dict, Optional, Type

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

from ai_traineree import DEVICE
from ai_traineree.agents import AgentBase
from ai_traineree.agents.agent_utils import soft_update
from ai_traineree.buffers.buffer_factory import BufferFactory
from ai_traineree.buffers.nstep import NStepBuffer
from ai_traineree.buffers.per import PERBuffer
from ai_traineree.loggers import DataLogger
from ai_traineree.networks import NetworkType, NetworkTypeClass
from ai_traineree.networks.heads import DuelingNet
from ai_traineree.types import AgentState, BufferState, DataSpace, NetworkState
from ai_traineree.types.experience import Experience
from ai_traineree.utils import to_numbers_seq, to_tensor


class DQNAgent(AgentBase):
    """Deep Q-Learning Network (DQN).

    The agent is not a vanilla DQN, although can be configured as such.
    The default config includes dual dueling nets and the priority experience buffer.
    Learning is also delayed by slowly copying to target nets (via tau parameter).
    Although NStep is implemented the default value is 1-step reward.

    There is also a specific implementation of the DQN called the Rainbow which differs
    to this implementation by working on the discrete space projection of the Q(s,a) function.
    """

    model = "DQN"

    def __init__(
        self,
        obs_space: DataSpace,
        action_space: DataSpace,
        network_fn: Callable[[], NetworkType] = None,
        network_class: Type[NetworkTypeClass] = None,
        state_transform: Optional[Callable] = None,
        reward_transform: Optional[Callable] = None,
        **kwargs
    ):
        """Initiates the DQN agent.

        Parameters:
            obs_space (DataSpace): Dataspace describing the input.
            action_space (DataSpace): Dataspace describing the output.
            network_fn (optional func): Function used to instantiate a network used by the agent.
            network_class (optional cls): Class of network that is instantiated with internal params to create network.
            state_transform (optional func): Function to transform (encode) state before used by the network.
            reward_transform (optional func): Function to transform reward before use.

        Keyword arguments:
            hidden_layers (tuple of ints): Shape of the hidden layers in fully connected network. Default: (64, 64).
            lr (float): Learning rate value. Default: 3e-4.
            gamma (float): Discount factor. Default: 0.99.
            tau (float): Soft-copy factor. Default: 0.002.
            update_freq (int): Number of steps between each learning step. Default 1.
            batch_size (int): Number of samples to use at each learning step. Default: 80.
            buffer_size (int): Number of most recent samples to keep in memory for learning. Default: 1e5.
            warm_up (int): Number of samples to observe before starting any learning step. Default: 0.
            number_updates (int): How many times to use learning step in the learning phase. Default: 1.
            max_grad_norm (float): Maximum norm of the gradient used in learning. Default: 10.
            using_double_q (bool): Whether to use Double Q Learning network. Default: True.
            n_steps (int): Number of lookahead steps when estimating reward. See :class:`NStepBuffer`. Default: 3.

        """
        super().__init__(**kwargs)

        self.device = self._register_param(kwargs, "device", DEVICE, update=True)
        self.obs_space = obs_space
        self.action_space = action_space

        self.lr = float(self._register_param(kwargs, "lr", 3e-4))  # Learning rate
        self.gamma = float(self._register_param(kwargs, "gamma", 0.99))  # Discount value
        self.tau = float(self._register_param(kwargs, "tau", 0.002))  # Soft update

        self.update_freq = int(self._register_param(kwargs, "update_freq", 1))
        self.batch_size = int(self._register_param(kwargs, "batch_size", 64, update=True))
        self.buffer_size = int(self._register_param(kwargs, "buffer_size", int(1e5), update=True))
        self.warm_up = int(self._register_param(kwargs, "warm_up", 0))
        self.number_updates = int(self._register_param(kwargs, "number_updates", 1))
        self.max_grad_norm = float(self._register_param(kwargs, "max_grad_norm", 10))

        self.iteration: int = 0
        self.buffer = PERBuffer(**kwargs)
        self.using_double_q = bool(self._register_param(kwargs, "using_double_q", True))

        self.n_steps = int(self._register_param(kwargs, "n_steps", 1))
        self.n_buffer = NStepBuffer(n_steps=self.n_steps, gamma=self.gamma)

        hidden_layers = to_numbers_seq(self._register_param(kwargs, "hidden_layers", (64, 64)))
        self.state_transform = state_transform if state_transform is not None else lambda x: x
        self.reward_transform = reward_transform if reward_transform is not None else lambda x: x
        action_feat = action_space.to_feature()

        if network_fn is not None:
            self.net = network_fn()
            self.target_net = network_fn()
        elif network_class is not None:
            self.net = network_class(obs_space.shape, action_feat, hidden_layers=hidden_layers, device=self.device)
            self.target_net = network_class(
                obs_space.shape, action_feat, hidden_layers=hidden_layers, device=self.device
            )
        else:
            self.net = DuelingNet(obs_space.shape, action_feat, hidden_layers=hidden_layers, device=self.device)
            self.target_net = DuelingNet(obs_space.shape, action_feat, hidden_layers=hidden_layers, device=self.device)
        self.optimizer = optim.Adam(self.net.parameters(), lr=self.lr)
        self._loss: float = float("nan")

    @property
    def loss(self) -> Dict[str, float]:
        return {"loss": self._loss}

    @loss.setter
    def loss(self, value):
        if isinstance(value, dict):
            value = value["loss"]
        self._loss = value

    def __eq__(self, o: object) -> bool:
        return (
            super().__eq__(o)
            and isinstance(o, type(self))
            and self._config == o._config
            and self.buffer == o.buffer
            and self.n_buffer == o.n_buffer
            and self.get_network_state() == o.get_network_state()
        )

    def reset(self):
        self.net.reset_parameters()
        self.target_net.reset_parameters()
        self.optimizer = optim.Adam(self.net.parameters(), lr=self.lr)

    def step(self, exp: Experience) -> None:
        """Letting the agent to take a step.

        On some steps the agent will initiate learning step. This is dependent on
        the `update_freq` value.

        Parameters:
            obs (ObservationType): Observation.
            action (int): Discrete action associated with observation.
            reward (float): Reward obtained for taking action at state.
            next_obs (ObservationType): Observation in a state where the action took.
            done: (bool) Whether in terminal (end of episode) state.

        """
        assert isinstance(exp.action, int), "DQN expects discrete actions (int)"
        self.iteration += 1
        t_obs = to_tensor(self.state_transform(exp.obs)).float().to("cpu")
        t_next_obs = to_tensor(self.state_transform(exp.next_obs)).float().to("cpu")
        reward = self.reward_transform(exp.reward)

        # Delay adding to buffer to account for n_steps (particularly the reward)
        self.n_buffer.add(
            obs=t_obs.numpy(),
            action=[int(exp.action)],
            reward=[reward],
            done=[exp.done],
            next_obs=t_next_obs.numpy(),
        )
        if not self.n_buffer.available:
            return

        self.buffer.add(**self.n_buffer.get().get_dict())

        if self.iteration < self.warm_up:
            return

        if len(self.buffer) >= self.batch_size and (self.iteration % self.update_freq) == 0:
            for _ in range(self.number_updates):
                self.learn(self.buffer.sample())

            # Update networks only once - sync local & target
            soft_update(self.target_net, self.net, self.tau)

    def act(self, experience: Experience, eps: float = 0.0) -> Experience:
        """Returns actions for given obs as per current policy.

        Parameters:
            experience (Experience): current observation
            eps (optional float): epsilon, for epsilon-greedy action selection. Default 0.

        Returns:
            Categorical value for the action.

        """
        # Epsilon-greedy action selection
        if self._rng.random() < eps:
            action = self._rng.randint(self.action_space.low, self.action_space.high)
            return experience.update(action=action)

        t_obs = to_tensor(self.state_transform(experience.obs)).float()
        t_obs = t_obs.unsqueeze(0).to(self.device)
        action_values = self.net.act(t_obs)
        action = int(torch.argmax(action_values.cpu()))
        return experience.update(action=action)

    def learn(self, experiences: Dict[str, list]) -> None:
        """Updates agent's networks based on provided experience.

        Parameters:
            experiences: Samples experiences from the experience buffer.

        """
        rewards = to_tensor(experiences["reward"]).type(torch.float32).to(self.device)
        dones = to_tensor(experiences["done"]).type(torch.int).to(self.device)
        obss = to_tensor(experiences["obs"]).type(torch.float32).to(self.device)
        next_obss = to_tensor(experiences["next_obs"]).type(torch.float32).to(self.device)
        actions = to_tensor(experiences["action"]).type(torch.long).to(self.device)

        with torch.no_grad():
            Q_targets_next = self.target_net.act(next_obss).detach()
            if self.using_double_q:
                _a = torch.argmax(self.net(next_obss), dim=-1).unsqueeze(-1)
                max_Q_targets_next = Q_targets_next.gather(1, _a)
            else:
                max_Q_targets_next = Q_targets_next.max(1)[0].unsqueeze(1)
        Q_targets = rewards + self.n_buffer.n_gammas[-1] * max_Q_targets_next * (1 - dones)
        Q_expected: torch.Tensor = self.net(obss).gather(1, actions)
        loss = F.mse_loss(Q_expected, Q_targets)

        self.optimizer.zero_grad()
        loss.backward()
        nn.utils.clip_grad_norm_(self.net.parameters(), self.max_grad_norm)
        self.optimizer.step()
        self._loss = float(loss.item())

        if hasattr(self.buffer, "priority_update"):
            error = Q_expected - Q_targets
            assert any(~torch.isnan(error))
            self.buffer.priority_update(experiences["index"], error.abs())

    def state_dict(self) -> Dict[str, dict]:
        """Describes agent's networks.

        Returns:
            state: (dict) Provides actors and critics states.

        """
        return {
            "net": self.net.state_dict(),
            "target_net": self.target_net.state_dict(),
        }

    def log_metrics(self, data_logger: DataLogger, step: int, full_log: bool = False):
        """Uses provided DataLogger to provide agent's metrics.

        Parameters:
            data_logger (DataLogger): Instance of the SummaryView, e.g. torch.utils.tensorboard.SummaryWritter.
            step (int): Ordering value, e.g. episode number.
            full_log (bool): Whether to all available information. Useful to log with lesser frequency.
        """
        data_logger.log_value("loss/agent", self._loss, step)

    def get_state(self) -> AgentState:
        """Provides agent's internal state."""
        return AgentState(
            model=self.model,
            obs_space=self.obs_space,
            action_space=self.action_space,
            config=self._config,
            buffer=copy.deepcopy(self.buffer.get_state()),
            network=copy.deepcopy(self.get_network_state()),
        )

    def get_network_state(self) -> NetworkState:
        return NetworkState(net=dict(net=self.net.state_dict(), target_net=self.target_net.state_dict()))

    @staticmethod
    def from_state(state: AgentState) -> AgentBase:
        config = copy.copy(state.config)
        config.update({"obs_space": state.obs_space, "action_space": state.action_space})
        agent = DQNAgent(**config)
        if state.network is not None:
            agent.set_network(state.network)
        if state.buffer is not None:
            agent.set_buffer(state.buffer)
        return agent

    def set_buffer(self, buffer_state: BufferState) -> None:
        self.buffer = BufferFactory.from_state(buffer_state)

    def set_network(self, network_state: NetworkState) -> None:
        self.net.load_state_dict(network_state.net["net"])
        self.target_net.load_state_dict(network_state.net["target_net"])

    def save_state(self, path: str):
        """Saves agent's state into a file.

        Parameters:
            path: String path where to write the state.

        """
        agent_state = self.get_state()
        torch.save(agent_state, path)

    def load_state(self, *, path: Optional[str] = None, state: Optional[AgentState] = None) -> None:
        """Loads state from a file under provided path.

        Parameters:
            path: String path indicating where the state is stored.

        """
        if path is None and state is None:
            raise ValueError("Either `path` or `state` must be provided to load agent's state.")
        if path is not None:
            state = torch.load(path)

        # Populate agent
        agent_state = state.agent
        self._config = agent_state.config
        self.__dict__.update(**self._config)

        # Populate network
        network_state = state.network
        self.net.load_state_dict(network_state.net["net"])
        self.target_net.load_state_dict(network_state.net["target_net"])
        self.buffer = PERBuffer(**self._config)

    def save_buffer(self, path: str) -> None:
        """Saves data from the buffer into a file under provided path.

        Parameters:
            path: String path where to write the buffer.

        """
        import json

        dump = self.buffer.dump_buffer(serialize=True)
        with open(path, "w") as f:
            json.dump(dump, f)

    def load_buffer(self, path: str) -> None:
        """Loads data into the buffer from provided file path.

        Parameters:
            path: String path indicating where the buffer is stored.

        """
        import json

        with open(path, "r") as f:
            buffer_dump = json.load(f)
        self.buffer.load_buffer(buffer_dump)



import json
import logging
import os
import sys
import time
from collections import deque
from pathlib import Path
from typing import Any, Iterable, List, Optional, Tuple

from ai_traineree.agents import AgentBase
from ai_traineree.loggers import DataLogger
from ai_traineree.types import RewardType, TaskType
from ai_traineree.types.experience import Experience
from ai_traineree.utils import save_gif

FRAMES_PER_SEC = 45

logging.basicConfig(
    stream=sys.stdout, level=logging.INFO, format="%(asctime)s %(levelname)-8s %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
)


class EnvRunner:
    """
    EnvRunner, short for Environment Runner, is meant to be used as module that runs your experiments.
    It's expected that the environments are wrapped in a Task which has typical step and act methods.
    The agent can be any agent which *makes sense* as there aren't any checks like whether the output is discrete.

    Examples:
        >>> env_runner = EnvRunner(task, agent)
        >>> env_runner.run()
    """

    logger = logging.getLogger("EnvRunner")

    def __init__(self, task: TaskType, agent: AgentBase, max_iterations: int = int(1e5), **kwargs):
        """
        Expects the environment to come as the TaskType and the agent as the AgentBase.

        Keyword Arguments:
            window_len (int): Length of the score averaging window. Default 50.
            data_logger: An instance of Data Logger, e.g. TensorboardLogger.
            logger_level: Logging level. Default: logging.INFO.

        """
        self.task = task
        self.agent = agent
        self.max_iterations = max_iterations
        self.model_path = f"{task.name}_{agent.model}"
        self.state_dir = "run_states"

        self.episode = 0
        self.iteration = 0
        self.all_scores = []
        self.all_iterations = []
        self.window_len = kwargs.get("window_len", 50)
        self.scores_window = deque(maxlen=self.window_len)
        self.__images = []

        self.logger.setLevel(kwargs.get("logger_level", logging.INFO))
        self.data_logger: Optional[DataLogger] = kwargs.get("data_logger")
        if self.data_logger:
            self.logger.info("DataLogger: %s", str(self.data_logger))
            self.data_logger.set_hparams(hparams=self.agent.hparams)

        self._debug_log: bool = bool(kwargs.get("debug_log", False))
        self._exp: List[Tuple[int, Experience]] = []
        self._actions: List[Any] = []
        self._states: List[Any] = []
        self._rewards: List[Any] = []
        self._dones: List[Any] = []
        self._noises: List[Any] = []

        self.seed(kwargs.get("seed"))

    def __str__(self) -> str:
        return f"EnvRunner<{self.task.name}, {self.agent.model}>"

    def seed(self, seed):
        if isinstance(seed, (int, float)):
            self.agent.seed(seed)
            self.task.seed(seed)

    def reset(self):
        """Resets the EnvRunner. The task env and the agent are preserved."""
        self.episode = 0
        self.all_scores = []
        self.all_iterations = []
        self.scores_window = deque(maxlen=self.window_len)

    def interact_episode(
        self,
        train: bool = False,
        eps: float = 0,
        max_iterations: Optional[int] = None,
        render: bool = False,
        render_gif: bool = False,
        log_interaction_freq: Optional[int] = 10,
        full_log_interaction_freq: Optional[int] = 1000,
    ) -> Tuple[RewardType, int]:
        score = 0
        obs = self.task.reset()
        iterations = 0
        max_iterations = max_iterations if max_iterations is not None else self.max_iterations
        done = False

        # Only gifs require keeping (S, A, R) list
        if render_gif:
            self.__images = []

        self.agent.train = train

        while iterations < max_iterations and not done:
            iterations += 1
            self.iteration += 1
            if render:
                self.task.render("human")
                time.sleep(1.0 / FRAMES_PER_SEC)

            experience = Experience(obs=obs)
            experience = self.agent.act(experience, eps)
            assert experience.get("action") is not None, "Act method should update action on experience"
            action = experience.action

            next_obs, reward, done, _ = self.task.step(action)
            self._rewards.append((self.iteration, reward))

            if self._debug_log:
                self._exp.append((self.iteration, experience))
                self._actions.append((self.iteration, action))
                self._states.append((self.iteration, obs))
                self._dones.append((self.iteration, done))
                noise = experience.get("noise")
                if noise is not None:
                    self._noises.append((self.iteration, noise))

            score += float(reward)
            if render_gif:
                # OpenAI gym still renders the image to the screen even though it shouldn't. Eh.
                img = self.task.render(mode="rgb_array")
                self.__images.append(img)

            experience.update(action=action, reward=reward, next_obs=next_obs, done=done)
            self.agent.step(experience)

            # Plot interactions every `log_interaction_freq` iterations.
            # Plot full log (including weights) every `full_log_interaction_freq` iterations.
            if (log_interaction_freq and (iterations % log_interaction_freq) == 0) or (
                full_log_interaction_freq and (self.iteration % full_log_interaction_freq) == 0
            ):
                full_log = full_log_interaction_freq and (iterations % full_log_interaction_freq) == 0
                self.log_data_interaction(full_log=full_log)

            # n -> n+1  => S(n) <- S(n+1)
            obs = next_obs

        if render_gif and len(self.__images):
            gif_path = "gifs/{}_e{}.gif".format(self.model_path, str(self.episode))
            save_gif(gif_path, self.__images)
            self.__images = []

        return score, iterations

    def run(
        self,
        reward_goal: float = 100.0,
        max_episodes: int = 2000,
        test_every: int = 10,
        eps_start: float = 1.0,
        eps_end: float = 0.01,
        eps_decay: float = 0.995,
        log_episode_freq: int = 1,
        log_interaction_freq: int = 10,
        gif_every_episodes: Optional[int] = None,
        checkpoint_every: Optional[int] = 200,
        force_new: bool = False,
    ) -> List[float]:
        """
        Evaluates the agent in the environment.
        The evaluation will stop when the agent reaches the `reward_goal` in the averaged last `self.window_len`, or
        when the number of episodes reaches the `max_episodes`.

        To help debugging one can set the `gif_every_episodes` to a positive integer which relates to how often a gif
        of the episode evaluation is written to the disk.

        Every `checkpoint_every` (default: 200) iterations the Runner will store current state of the runner
        and the agent. These states can be used to resume previous run. By default the runner checks whether
        there is ongoing run for the combination of the environment and the agent.

        Parameters:
            reward_goal: Goal to achieve on the average reward.
            max_episode: After how many episodes to stop regardless of the score.
            test_every: Number of episodes between agent test run (without learning). Default: 10.
            eps_start: Epsilon-greedy starting value.
            eps_end: Epislon-greeedy lowest value.
            eps_decay: Epislon-greedy decay value, eps[i+1] = eps[i] * eps_decay.
            log_episode_freq: Number of episodes between state logging.
            gif_every_episodes: Number of episodes between storing last episode as a gif.
            checkpoint_every: Number of episodes between storing the whole state, so that
                in case of failure it can be safely resumed from it.
            force_new: Flag whether to resume from previously stored state (False), or to
                start learning from a clean state (True).

        Returns:
            All obtained scores from all episodes.

        """
        self.epsilon = eps_start
        self.reset()
        if not force_new:
            self.load_state(file_prefix=self.model_path)
        mean_scores = []
        epsilons = []

        while self.episode < max_episodes:
            self.episode += 1
            train_agent = (self.episode % test_every) != 0
            render_gif = gif_every_episodes is not None and (self.episode % gif_every_episodes) == 0
            eps = self.epsilon if train_agent else 0

            score, iterations = self.interact_episode(
                train=train_agent, eps=eps, render_gif=render_gif, log_interaction_freq=log_interaction_freq
            )

            if not train_agent:
                self.scores_window.append(score)
                mean_scores.append(sum(self.scores_window) / len(self.scores_window))

            self.all_iterations.append(iterations)
            self.all_scores.append(score)

            epsilons.append(eps)

            self.epsilon = max(eps_end, eps_decay * self.epsilon)

            if self.episode % log_episode_freq == 0:
                last_episodes = [self.episode - i for i in range(log_episode_freq)[::-1]]
                self.info(
                    train_mode=train_agent,
                    episodes=last_episodes,
                    iterations=self.all_iterations[-log_episode_freq:],
                    scores=self.all_scores[-log_episode_freq:],
                    mean_scores=mean_scores[-log_episode_freq:],
                    epsilons=epsilons[-log_episode_freq:],
                    loss=self.agent.loss,
                )

            if len(mean_scores) and mean_scores[-1] >= reward_goal and len(self.scores_window) == self.window_len:
                print(f"Environment solved after {self.episode} episodes!\tAverage Score: {mean_scores[-1]:.2f}")
                self.save_state(self.model_path)
                self.agent.save_state(f"{self.model_path}_agent.net")
                break

            if checkpoint_every is not None and self.episode % checkpoint_every == 0:
                self.save_state(self.model_path)
        # END training

        # Store hyper parameters and experiment metrics in logger so that it's easier to compare runs
        if self.data_logger:
            end_metrics = {
                "hparam/total_iterations": sum(self.all_iterations),
                "hparam/total_episodes": len(self.all_iterations),
                "hparam/score": mean_scores[-1],
            }
            hparams = {**self.agent.hparams, **end_metrics}
            self.data_logger.set_hparams(hparams=hparams)

        return self.all_scores

    def info(self, **kwargs):
        """
        Writes out current state into provided loggers.
        Writting to stdout is done through Python's logger, whereas all metrics
        are supposed to be handled via DataLogger. Currently supported are Tensorboard
        and Neptune (neptune.ai). To use one of these `data_logger` is expected.
        """
        if self.data_logger is not None:
            self.log_episode_metrics(**kwargs)
            self.log_data_interaction(**kwargs)
        if self.logger is not None:
            self.log_logger(**kwargs)

    def log_logger(self, **kwargs):
        """Writes out env logs via logger (either stdout or a file)."""
        if kwargs is None:
            return

        line_chunks = []
        episodes = kwargs.get("episodes")
        if episodes is not None:
            is_train = kwargs.get("train_mode", True)
            test_mark = "" if is_train else "T"
            line_chunks += [f"Episode {episodes[-1]}{test_mark}"]

        scores = kwargs.get("scores")
        if scores is not None:
            line_chunks += [f"Current Score: {scores[-1]:.2f}"]

        iterations = kwargs.get("iterations")
        if iterations is not None:
            line_chunks += [f"Iter: {iterations[-1]}"]

        mean_scores = kwargs.get("mean_scores")
        if mean_scores is not None and len(mean_scores):
            line_chunks += [f"Average Score: {mean_scores[-1]:.2f}"]

        epsilons = kwargs.get("epsilons")
        if epsilons is not None:
            line_chunks += [f"Epsilon: {epsilons[-1]:5.3f}"]

        loss = kwargs.get("loss", {})
        if loss is not None:
            line_chunks += [f"{loss_name.capitalize()}: {loss_value:10.4f}" for (loss_name, loss_value) in loss.items()]

        line = ";\t".join(line_chunks)
        self.logger.info(line)

    def log_episode_metrics(self, **kwargs):
        """Uses DataLogger, e.g. TensorboardLogger, to store env metrics."""
        episodes: List[int] = kwargs.get("episodes", [])
        for episode, epsilon in zip(episodes, kwargs.get("epsilons", [])):
            self.data_logger.log_value("episode/epsilon", epsilon, episode)

        for episode, mean_score in zip(episodes, kwargs.get("mean_scores", [])):
            self.data_logger.log_value("episode/avg_score", mean_score, episode)

        for episode, score in zip(episodes, kwargs.get("scores", [])):
            self.data_logger.log_value("episode/score", score, episode)

        for episode, iteration in zip(episodes, kwargs.get("iterations", [])):
            self.data_logger.log_value("episode/iterations", iteration, episode)

    def log_data_interaction(self, **kwargs):
        if self.data_logger is None:
            return

        if hasattr(self.agent, "log_metrics"):
            self.agent.log_metrics(self.data_logger, self.iteration, full_log=kwargs.get("full_log", False))
        else:
            for loss_name, loss_value in kwargs.get("loss", {}).items():
                self.data_logger.log_value(f"loss/{loss_name}", loss_value, self.iteration)

        while self._debug_log and self._exp:
            step, exp = self._exp.pop(0)
            noise_params = exp.get("noise_params")
            if not noise_params:
                continue
            self.data_logger.log_values_dict("env/noise_params", {str(i): a for i, a in enumerate(noise_params)}, step)

        while self._debug_log and self._states:
            step, states = self._states.pop(0)
            states = states if isinstance(states, Iterable) else [states]
            self.data_logger.log_values_dict("env/states", {str(i): a for i, a in enumerate(states)}, step)

        while self._debug_log and self._actions:
            step, actions = self._actions.pop(0)
            actions = actions if isinstance(actions, Iterable) else [actions]
            self.data_logger.log_values_dict("env/action", {str(i): a for i, a in enumerate(actions)}, step)

        while self._debug_log and self._noises:
            step, noises = self._noises.pop(0)
            noises = noises if isinstance(noises, Iterable) else [noises]
            self.data_logger.log_values_dict("env/noise", {str(i): a for i, a in enumerate(noises)}, step)

        while self._debug_log and self._rewards:
            step, rewards = self._rewards.pop(0)
            self.data_logger.log_value("env/reward", float(rewards), step)

        while self._debug_log and self._dones:
            step, dones = self._dones.pop(0)
            self.data_logger.log_value("env/done", int(dones), step)

    def save_state(self, state_name: str):
        """Saves the current state of the runner and the agent.

        Files are stored with appended episode number.
        Agents are saved with their internal saving mechanism.
        """
        state = {
            "tot_iterations": sum(self.all_iterations),
            "episode": self.episode,
            "epsilon": self.epsilon,
            "score": self.all_scores[-1],
            "average_score": sum(self.scores_window) / len(self.scores_window),
            "loss": self.agent.loss,
        }

        Path(self.state_dir).mkdir(parents=True, exist_ok=True)
        self.agent.save_state(f"{self.state_dir}/{state_name}_e{self.episode}.agent")
        with open(f"{self.state_dir}/{state_name}_e{self.episode}.json", "w") as f:
            json.dump(state, f)

    def load_state(self, file_prefix: str):
        """
        Loads state with the highest episode value for given agent and environment.
        """
        try:
            state_files = list(
                filter(lambda f: f.startswith(file_prefix) and f.endswith("json"), os.listdir(self.state_dir))
            )
            recent_episode_num = max([int(f[f.index("_e") + 2 : f.index(".")]) for f in state_files])
            state_name = [n for n in state_files if n.endswith(f"_e{recent_episode_num}.json")][0][:-5]
        except Exception:
            self.logger.warning("Couldn't load state. Forcing restart.")
            return

        self.logger.info("Loading saved state under: %s/%s.json", self.state_dir, state_name)
        with open(f"{self.state_dir}/{state_name}.json", "r") as f:
            state = json.load(f)
        self.episode = state.get("episode")
        self.epsilon = state.get("epsilon")

        self.all_scores.append(state.get("score"))
        self.all_iterations = []

        avg_score = state.get("average_score")
        for _ in range(min(self.window_len, self.episode)):
            self.scores_window.append(avg_score)

        self.logger.info("Loading saved agent state: %s/%s.agent", self.state_dir, state_name)
        self.agent.load_state(path=f"{self.state_dir}/{state_name}.agent")
        self.agent.loss = state.get("loss", 0)



from functools import reduce
from math import sqrt
from typing import Any, Optional, Sequence, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F

from ai_traineree.networks import NetworkType
from ai_traineree.types import FeatureType


def hidden_init(layer: nn.Module):
    fan_in = layer.weight.data.size()[0]  # type: ignore
    lim = 1.0 / sqrt(fan_in)
    return (-lim, lim)


def layer_init(layer: nn.Module, range_value: Optional[Tuple[float, float]] = None, remove_mean=True):
    if not (isinstance(layer, nn.Conv2d) or isinstance(layer, nn.Linear)):
        return
    if range_value is not None:
        layer.weight.data.uniform_(*range_value)  # type: ignore
    if remove_mean:
        layer.weight.data -= layer.weight.data.mean()

    nn.init.xavier_uniform_(layer.weight)


class ScaleNet(NetworkType):
    def __init__(self, scale: Union[float, int]) -> None:
        super(ScaleNet, self).__init__()
        self.scale = scale

    def forward(self, x):
        return x * self.scale


class ConvNet(NetworkType):
    def __init__(self, input_dim: Sequence[int], **kwargs):
        """Convolution Network.

        Constructs a layered network over torch.nn.Conv2D. Number of layers is set based on `hidden_layers` argument.
        To update other arguments, e.g. kernel_size or bias, pass either a single value or a tuple of the same
        length as `hidden_layers`.

        Quick reminder from the PyTorch doc (https://pytorch.org/docs/stable/generated/torch.nn.Conv2d.html).

        Keyword Arguments:
            in_channels (int): Number of channels in the input image
            hidden_layers (tuple of ints): Number of channels in each hidden layer
            kernel_size (int or tuple): Size of the convolving kernel
            stride (int or tuple, optional): Stride of the convolution. Default: 1
            padding (int or tuple, optional): Zero-padding added to both sides of the input. Default: 0
            padding_mode (string, optional): 'zeros', 'reflect', 'replicate' or 'circular'. Default: 'zeros'
            dilation (int or tuple, optional): Spacing between kernel elements. Default: 1
            groups (int, optional): Number of blocked connections from input channels to output channels. Default: 1
            bias (bool, optional): If True, adds a learnable bias to the output. Default: True

        Examples:
            >>> config = {"hidden_layers": (300, 200, 100), "kernel_size": 6, "gate": F.relu}
            >>> net = ConvNet(input_dim=(10, 10, 3), **config)
            >>> config = {"hidden_layers": (64, 32, 64), "kernel_size": (3, 4, 3), padding: 2, "gate": F.relu}
            >>> net = ConvNet(input_dim=(20, 10, 1), **config)

        """
        super(ConvNet, self).__init__()

        # input_dim = (num_layers, x_img, y_img, channels)
        hidden_layers = kwargs.get("hidden_layers", (20, 20))
        num_layers = [input_dim[0]] + list(hidden_layers)

        gate = kwargs.get("gate", nn.ReLU)
        max_pool_sizes = self._expand_to_seq(kwargs.get("max_pool_size", 2), len(hidden_layers))
        kernel_sizes = self._expand_to_seq(kwargs.get("kernel_size", 3), len(hidden_layers))
        strides = self._expand_to_seq(kwargs.get("stride", 1), len(hidden_layers))
        paddings = self._expand_to_seq(kwargs.get("padding", 0), len(hidden_layers))
        dilations = self._expand_to_seq(kwargs.get("dilation", 1), len(hidden_layers))
        biases = self._expand_to_seq(kwargs.get("bias", True), len(hidden_layers))

        layers = []
        for layer_idx in range(len(hidden_layers)):
            layers.append(
                nn.Conv2d(
                    num_layers[layer_idx],
                    num_layers[layer_idx + 1],
                    kernel_size=kernel_sizes[layer_idx],
                    stride=strides[layer_idx],
                    padding=paddings[layer_idx],
                    dilation=dilations[layer_idx],
                    bias=biases[layer_idx],
                )
            )

            if max_pool_sizes[layer_idx] > 1:
                layers.append(nn.MaxPool2d(max_pool_sizes[layer_idx]))

            if gate is not None:
                layers.append(gate())

        self.layers = nn.ModuleList(layers)
        self.reset_parameters()

        self.input_dim = input_dim
        self.device = kwargs.get("device")
        self.to(self.device)

    @staticmethod
    def _expand_to_seq(o: Union[Any, Sequence[Any]], size) -> Sequence[Any]:
        return o if isinstance(o, Sequence) else (o,) * size

    @property
    def output_size(self):
        return reduce(lambda a, b: a * b, self._calculate_output_size(self.input_dim, self.layers))

    @torch.no_grad()
    def _calculate_output_size(self, input_dim: Sequence[int], layers) -> Sequence[int]:
        test_tensor = torch.zeros((1,) + tuple(input_dim)).to(self.device)
        out = self.forward(test_tensor)
        return out.shape

    def reset_parameters(self):
        self.layers.apply(layer_init)

    def forward(self, x):
        for layer in self.layers:
            x = layer(x)
        return x


class FcNet(NetworkType):
    """
    For the activation layer we use tanh by default which was observed to be much better, e.g. compared to ReLU,
    for policy networks [1]. The last gate, however, might be changed depending on the actual task.


    References
        .. [1] "What Matters In On-Policy Reinforcement Learning? A Large-Scale Empirical Study"
            by M. Andrychowicz et al. (2020). Link: https://arxiv.org/abs/2006.05990

    """

    def __init__(
        self,
        in_features: FeatureType,
        out_features: FeatureType,
        hidden_layers: Optional[Sequence[int]] = (200, 100),
        last_layer_range=(-3e-4, 3e-4),
        bias: bool = True,
        **kwargs,
    ):
        """Fully Connected network with default APIs.

        Parameters:
            in_features (sequence of ints): Shape of the input.
            out_features (sequence of ints): Shape of the output.
            hidden_layers: Shape of the hidden layers. If None, then the output is directly computed from the input.
            last_layer_range: The range for the uniform distribution that initiates the last layer.

        Keyword arguments:
            gate (optional torch.nn.layer): Activation function for each layer, expect the last. Default: torch.tanh.
            gate_out (optional torch.nn.layer): Activation function after the last layer. Default: Identity layer.
            device (torch.devce or str): Device where to allocate memory. CPU or CUDA.


        """
        super(FcNet, self).__init__()
        assert len(in_features) == 1, "Expected only one dimension"
        assert len(out_features) == 1, "Expected only one dimension"

        self.in_features = tuple(in_features)
        self.out_features = tuple(out_features)
        num_layers = tuple(hidden_layers) if hidden_layers is not None else tuple()
        num_layers = self.in_features + num_layers + self.out_features
        layers = [nn.Linear(dim_in, dim_out, bias=bias) for dim_in, dim_out in zip(num_layers[:-1], num_layers[1:])]

        self.last_layer_range = last_layer_range
        self.layers = nn.ModuleList(layers)
        self.reset_parameters()

        self.gate = kwargs.get("gate", torch.tanh)
        self.gate_out = kwargs.get("gate_out", nn.Identity())
        self.to(kwargs.get("device"))

    def reset_parameters(self):
        for layer in self.layers[:-1]:
            layer_init(layer, hidden_init(layer), remove_mean=True)
        layer_init(self.layers[-1], self.last_layer_range, remove_mean=True)

    def forward(self, x):
        for layer in self.layers[:-1]:
            x = self.gate(layer(x))
        return self.gate_out(self.layers[-1](x))


###
# In most cases, the default ActorBody can be associated with a fully connected network.
# The alias is only for convinience and, hopefully, better understanding of some algorithms.
###
ActorBody = FcNet


class CriticBody(NetworkType):
    """Extension of the FcNet which includes actions.

    Mainly used to estimate the state-action value function in actor-critic agents.
    Actions are included (by default) in the first hidden layer (changeable).

    Since the main purpose for this is value function estimation the output is a single value.
    """

    def __init__(
        self,
        in_features: FeatureType,
        inj_action_size: int,
        out_features: FeatureType = (1,),
        hidden_layers: Optional[Sequence[int]] = (100, 100),
        inj_actions_layer: int = 1,
        **kwargs,
    ):
        """
        Parameters:
            in_features (tuple of ints): Dimension of the input features.
            inj_action_size (int): Dimension of the action vector that is injected into `inj_action_layer`.
            out_features (tuple of ints): Dimension of critic's action. Default: (1,).
            hidden_layers (tuple of ints): Shape of the hidden layers. Default: (100, 100).
            inj_action_layer (int): An index for the layer that will have `actions` injected as an additional input.
                By default that's a first hidden layer, i.e. (state) -> (out + actions) -> (out) ... -> (output).
                Default: 1.

        Keyword arguments:
            bias (bool): Whether to include bias in network's architecture. Default: True.
            gate (callable): Activation function for each layer, expect the last. Default: Identity layer.
            gate_out (callable): Activation function after the last layer. Default: Identity layer.
            device: Device where to allocate memory. CPU or CUDA. Default CUDA if available.

        """
        super().__init__()

        self.in_features = tuple(in_features)
        self.out_features = tuple(out_features)
        num_layers = tuple(hidden_layers) if hidden_layers is not None else tuple()
        num_layers = self.in_features + num_layers + self.out_features
        self.actions_layer = inj_actions_layer
        if not (0 <= inj_actions_layer < len(num_layers)):
            raise ValueError("Action layer needs to be within the network")

        bias = bool(kwargs.get("bias", True))
        layers = []
        for in_idx in range(len(num_layers) - 1):
            in_dim, out_dim = num_layers[in_idx], num_layers[in_idx + 1]
            if in_idx == inj_actions_layer:  # Injects `actions` into specified (default: 2nd) layer of the Critic
                in_dim += inj_action_size
            layers.append(nn.Linear(in_dim, out_dim, bias=bias))

        self.layers = nn.ModuleList(layers)
        self.reset_parameters()

        self.gate = kwargs.get("gate", nn.Identity())
        self.gate_out = kwargs.get("gate_out", nn.Identity())
        self.to(kwargs.get("device"))

    def reset_parameters(self):
        for layer in self.layers:
            layer_init(layer, hidden_init(layer))

    def forward(self, x, actions):
        for idx, layer in enumerate(self.layers):
            if idx == self.actions_layer:
                x = layer(torch.cat((x, actions.float()), dim=-1))
            else:
                x = layer(x)

            if idx < len(self.layers) - 1:
                x = self.gate(x)
            else:
                x = self.gate_out(x)
        return x


class NoisyLayer(nn.Module):
    def __init__(
        self, in_features: FeatureType, out_features: FeatureType, sigma: float = 0.4, factorised: bool = True
    ):
        """
        A linear layer with added noise perturbations in training as described in [1].
        For a fully connected network of NoisyLayers see :class:`.NoisyNet`.

        Parameters:
            in_features (tuple ints): Dimension of the input.
            out_features (tuple ints): Dimension of the output.
            sigma (float): Used to intiated noise distribution. Default: 0.4.
            factorised: Whether to use independent Gaussian (False) or Factorised Gaussian (True) noise.
                Suggested [1] for DQN and Duelling nets to use factorised as it's quicker.

        References:
            .. [1] "Noisy Networks for Exploration" by Fortunato et al. (ICLR 2018), https://arxiv.org/abs/1706.10295.

        """
        super(NoisyLayer, self).__init__()
        assert len(in_features) == 1, "Expected only one dimension"
        assert len(out_features) == 1, "Expected only one dimension"

        self.in_features = in_features
        self.out_features = out_features
        self.sigma_0 = sigma
        self.factorised = factorised

        self.weight_mu = nn.Parameter(torch.zeros((out_features[0], in_features[0])))
        self.weight_sigma = nn.Parameter(torch.zeros((out_features[0], in_features[0])))

        self.bias_mu = nn.Parameter(torch.zeros(out_features[0]))
        self.bias_sigma = nn.Parameter(torch.zeros(out_features[0]))

        self.register_buffer("weight_eps", torch.zeros((out_features[0], in_features[0])))
        self.register_buffer("bias_eps", torch.zeros(out_features[0]))

        self.bias_noise = torch.zeros(out_features[0])
        if factorised:
            self.weight_noise = torch.zeros(in_features[0])
        else:
            self.weight_noise = torch.zeros(out_features[0], in_features[0])

        self.reset_parameters()
        self.reset_noise()

    def forward(self, x) -> torch.Tensor:
        weight = self.weight_mu
        bias = self.bias_mu
        if self.training:
            weight = weight.add(self.weight_sigma.mul(self.weight_eps))
            bias = bias.add(self.bias_sigma.mul(self.bias_eps))

        return F.linear(x, weight, bias)

    def reset_parameters(self) -> None:
        if self.factorised:
            bound = sqrt(1.0 / self.in_features[0])
            sigma = self.sigma_0 * bound
        else:
            bound = sqrt(3.0 / self.in_features[0])
            sigma = 0.017  # Yes, that's correct. [1]

        self.weight_mu.data.uniform_(-bound, bound)
        self.weight_sigma.data.fill_(sigma)

        self.bias_mu.data.uniform_(-bound, bound)
        self.bias_sigma.data.fill_(sigma)

    def reset_noise(self):
        self.bias_noise.normal_(std=self.sigma_0)
        self.weight_noise.normal_(std=self.sigma_0)

        if self.factorised:
            # eps_i = ~P(b_size),  eps_j = ~P(w_size)
            # eps_b = f(eps_i)
            # eps_w = f(eps_i) x f(eps_j)
            f_weight_eps = self.noise_function(self.weight_noise)
            f_bias_eps = self.noise_function(self.bias_noise)
            self.weight_eps.copy_(f_bias_eps.outer(f_weight_eps))
            self.bias_eps.copy_(f_bias_eps)
        else:
            self.weight_eps.copy_(self.weight_noise.data)
            self.bias_eps.copy_(self.bias_noise.data)

    @staticmethod
    def noise_function(x):
        return x.sign().mul_(x.abs().sqrt())


class NoisyNet(NetworkType):
    def __init__(
        self,
        in_features: FeatureType,
        out_features: FeatureType,
        hidden_layers: Optional[Sequence[int]] = (100, 100),
        sigma=0.4,
        factorised=True,
        **kwargs,
    ):
        """
        Parameters:
            in_features (tuple ints): Dimension of the input.
            out_features (tuple ints): Dimension of the output.
            hidden_layers (sequence ints): Sizes of latent layers. Size of sequence denotes number of hidden layers and
                values of the sequence are nodes per layer. If None is passed then the input goes straight to output.
                Default: (100, 100).
            sigma (float): Variance value for generating noise in noisy layers. Default: 0.4 per layer.
            factorised (bool): Whether to use independent Gaussian (False) or Factorised Gaussian (True) noise.
                Suggested [1] for DQN and Duelling nets to use factorised as it's quicker.

        Keyword arguments:
            gate (callable): Function to apply after each layer pass. For the best performance it is suggested
                to use non-linear functions such as tanh. Default: tanh.
            gate_out (callable): Function to apply on network's exit. Default: identity.
            device (str or torch.device): Whether and where to cast the network. Default is CUDA if available else cpu.

        References:
            .. [1] "Noisy Networks for Exploration" by Fortunato et al. (ICLR 2018), https://arxiv.org/abs/1706.10295.

        """
        super(NoisyNet, self).__init__()
        assert len(in_features) == 1, "Expected only one dimension"
        assert len(out_features) == 1, "Expected only one dimension"

        self.in_features = in_features
        self.out_features = out_features
        num_layers = list(hidden_layers) if hidden_layers is not None else []
        num_layers = [self.in_features[0]] + num_layers + [self.out_features[0]]
        layers = [
            NoisyLayer((dim_in,), (dim_out,), sigma=sigma, factorised=factorised)
            for dim_in, dim_out in zip(num_layers[:-1], num_layers[1:])
        ]
        self.layers = nn.ModuleList(layers)

        self.gate = kwargs.get("gate", torch.tanh)
        self.gate_out = kwargs.get("gate_out", nn.Identity())
        if not callable(self.gate) or not callable(self.gate_out):
            raise ValueError("Passed gate or gate_out is no callable and cannot be used as a function")

        self.to(device=kwargs.get("device", None))

    def reset_noise(self) -> None:
        for layer in self.layers:
            layer.reset_noise()

    def reset_parameters(self) -> None:
        for layer in self.layers:
            layer.reset_parameters()

    def forward(self, x) -> torch.Tensor:
        for layer in self.layers[:-1]:
            x = self.gate(layer(x))
        return self.gate_out(self.layers[-1](x))



import pylab as plt

from ai_traineree.agents.ddpg import DDPGAgent as DDPG
from ai_traineree.loggers.tensorboard_logger import TensorboardLogger
from ai_traineree.runners.env_runner import EnvRunner
from ai_traineree.tasks import GymTask
from ai_traineree.types import TaskType

seed = 123
data_logger = TensorboardLogger()
env_name = "LunarLanderContinuous-v2"
task: TaskType = GymTask(env_name)
task.seed(seed)

config = {
    "warm_up": 1000,
    "update_freq": 50,
    "hidden_layers": (100, 100),
    "noise_scale": 1.0,
    "noise_sigma": 1.0,
    "number_updates": 1,
    "buffer_size": 1e5,
    "batch_size": 200,
    "tau": 0.002,
    "max_grad_norm_actor": 2.0,
    "max_grad_norm_critic": 2.0,
}
agent = DDPG(task.obs_space, task.action_space, **config)
env_runner = EnvRunner(task, agent, data_logger=data_logger, debug_log=True)
scores = env_runner.run(
    reward_goal=80,
    max_episodes=3000,
    eps_start=1.0,
    eps_end=0.3,
    eps_decay=0.999,
    force_new=True,
    checkpoint_every=None,
)
# env_runner.interact_episode(eps=0, render=True)
data_logger.close()

# plot the scores
fig = plt.figure()
ax = fig.add_subplot(111)
plt.plot(range(len(scores)), scores)
plt.ylabel("Score")
plt.xlabel("Episode #")
plt.savefig(f"{env_name}.png", dpi=120)
plt.show()



"""
This example uses DummyAgent on CartPole gym environment.
Since Dummy Agent only returns random values and never learns
this example is purely for demonsitration purpose only.
"""
import numpy as np
import pylab as plt
import torch

from ai_traineree.agents.dummy import DummyAgent
from ai_traineree.loggers.tensorboard_logger import TensorboardLogger
from ai_traineree.runners.env_runner import EnvRunner
from ai_traineree.tasks import GymTask

seed = 32167
# torch.set_deterministic(True)
torch.manual_seed(seed)
data_logger = TensorboardLogger()

env_name = "CartPole-v1"
task = GymTask(env_name, seed=seed)
agent = DummyAgent(task.obs_space, task.action_space)
env_runner = EnvRunner(task, agent, data_logger=data_logger, seed=seed)

scores = env_runner.run(reward_goal=100, max_episodes=300, force_new=True)
env_runner.interact_episode(render=True)

# plot the scores
fig = plt.figure()
ax = fig.add_subplot(111)
plt.plot(np.arange(len(scores)), scores)
plt.ylabel("Score")
plt.xlabel("Episode #")
plt.savefig(f"{env_name}.png", dpi=120)
plt.show()



import numpy as np
import pylab as plt

from ai_traineree.agents.dqn import DQNAgent
from ai_traineree.runners.env_runner import EnvRunner
from ai_traineree.tasks import GymTask
from ai_traineree.types import TaskType

env_name = "LunarLander-v2"
task: TaskType = GymTask(env_name)
config = {"batch_size": 64}
agent = DQNAgent(task.obs_space, task.action_space, config=config)
env_runner = EnvRunner(task, agent)

env_runner.interact_episode(eps=0, render=True)
scores = env_runner.run(50, 800, eps_start=1.0, eps_end=0.05, eps_decay=0.995)
env_runner.interact_episode(eps=0, render=True)

# plot the scores
fig = plt.figure()
ax = fig.add_subplot(111)
plt.plot(np.arange(len(scores)), scores)
plt.ylabel("Score")
plt.xlabel("Episode #")
plt.savefig(f"{env_name}.png", dpi=120)
plt.show()



import torch

from ai_traineree.policies import MultivariateGaussianPolicy, MultivariateGaussianPolicySimple


def test_multi_gauss_simple_defaults():
    # Assign
    size = 5
    std_init = 0.5
    test_loc = torch.zeros((1, size))  # Shape: (1, 5)
    policy = MultivariateGaussianPolicySimple(size, std_init=std_init)

    # Act
    samples = policy(test_loc)

    # Assert
    assert policy.std_min == 0.0001
    assert policy.std_max == 2.0
    assert policy.param_dim == 1
    assert all(policy.std == torch.ones(size) * std_init)  # 1D
    assert samples.shape == (1, size)


def test_multi_gauss_simple_std_updates():
    # Assign
    size = 5
    std_init, std_min, std_max = 2, 0.01, 10
    test_loc = torch.arange(size).unsqueeze(0)  # Shape: (1, 5)
    test_std = torch.full((size,), std_init)  # Shape: (5,)
    test_cov_mat = torch.eye(size).unsqueeze(0) * std_init**2  # Shape: (1, 5, 5)
    policy = MultivariateGaussianPolicySimple(size, std_init=std_init, std_min=std_min, std_max=std_max)

    # Act
    samples = policy(test_loc)

    # Assert
    assert policy.std_min == std_min
    assert policy.std_max == std_max
    assert all(policy.std == test_std)  # 1D
    assert samples.shape == (1, size)


def test_multi_gauss_simple_statistic():
    """Compares statistical measures. Might (very rarely) fail."""
    # Assign
    size = 3
    batch_size = 3000
    expected_loc = torch.tensor([-2.0, 0, 2.0])  # Shape: (3,)
    expected_std = test_std = torch.tensor([0.1, 1, 2.0])
    policy = MultivariateGaussianPolicySimple(size, std_max=max(expected_std) * 2)
    policy.std.data = test_std
    test_loc = expected_loc.repeat((batch_size, 1))  # Shape: (3000, 3)

    # Act
    samples = policy(test_loc)

    # Assert
    assert samples.shape == (batch_size, size)
    assert torch.all(torch.isclose(samples.std(dim=0), expected_std, atol=0.1))  # +/- 0.1
    assert torch.all(torch.abs(samples.mean(dim=0) - expected_loc) < 0.2 * expected_std)


def test_multi_gauss():
    # Assign
    size = 3
    policy = MultivariateGaussianPolicy(size)
    loc = torch.zeros((1, size))  # Shape: (1, 3)
    std = torch.ones((1, size))  # Shape: (1, 3)
    x = torch.hstack((loc, std))  # Shape: (1, 6)
    test_cov_mat = torch.eye(size).unsqueeze(0)  # Shape: (1, 3, 3)

    # Act
    samples = policy(x)

    # Assert
    assert policy.param_dim == 2
    assert x.shape == (1, policy.param_dim * size)  # Shape: (1, 6)
    assert samples.shape == (size,)

    assert torch.all(policy._last_dist.loc == loc)
    assert torch.all(policy._last_dist.covariance_matrix == test_cov_mat)


if __name__ == "__main__":
    test_multi_gauss_simple_statistic()



from typing import Any, Dict, Optional

import jsons

from ai_traineree.utils import to_list

from .primitive import ActionType, DoneType, ObsType, RewardType


class Experience:
    """
    Basic data unit to hold information.

    It typically represents a one whole cycle of observation - action - reward.
    Data type used to store experiences in experience buffers.
    """

    whitelist = [
        "action",
        "reward",
        "done",
        "obs",
        "next_obs",
        "state",
        "next_state",
        "advantage",
        "logprob",
        "value",
        "noise",
        "noise_params",
        "priority",
        "index",
        "weight",
        "state_idx",
        "next_state_idx",
    ]
    obs: ObsType
    action: ActionType
    reward: RewardType
    done: DoneType
    next_obs: Optional[ObsType]
    state: ObsType
    next_state: ObsType

    def __init__(self, **kwargs):
        self.data = {}
        self.update(**kwargs)

    def __eq__(self, o: object) -> bool:
        return isinstance(o, Experience) and self.data == o.data

    def get(self, key: str):
        return self.data.get(key)

    def update(self, **kwargs):
        for (key, value) in kwargs.items():
            if key in Experience.whitelist:
                self.data[key] = value
                self.__dict__[key] = value  # TODO: Delete after checking that everything is updated
        return self

    def __add__(self, o_exp):
        return self.update(**o_exp.get_dict())

    def get_dict(self, serialize=False) -> Dict[str, Any]:
        if serialize:
            return {k: to_list(v) for (k, v) in self.data.items()}
        return self.data


def exprience_serialization(obj: Experience, **kwargs) -> Dict[str, Any]:
    # return {k: to_list(v) for (k, v) in obj.data.items() if v is not None}
    return {k: jsons.dumps(v) for (k, v) in obj.data.items()}


jsons.set_serializer(exprience_serialization, Experience)



import numpy as np
import pylab as plt
import sneks  # noqa
import torch.nn as nn

from ai_traineree.agents.dqn import DQNAgent
from ai_traineree.loggers.tensorboard_logger import TensorboardLogger
from ai_traineree.networks.bodies import ConvNet, FcNet
from ai_traineree.networks.heads import NetChainer
from ai_traineree.runners.env_runner import EnvRunner
from ai_traineree.tasks import GymTask


def running_mean(x, N):
    cumsum = np.cumsum(np.insert(x, 0, 0))
    return (cumsum[N:] - cumsum[:-N]) / float(N)


def state_transform(state: np.ndarray):
    return state[1:-1, 1:-1][None, ...]


def network_fn(state_dim, output_dim, device):
    conv_net = ConvNet(state_dim, hidden_layers=(40, 20), device=device)
    return NetChainer(
        net_classes=[
            conv_net,
            nn.Flatten(),
            FcNet((conv_net.output_size,), (output_dim,), hidden_layers=(100, 100), device=device),
        ]
    )


data_logger = TensorboardLogger()
env_name = "hungrysnek-raw-16-v1"
task = GymTask(env_name, state_transform=state_transform)

input_obs_shape = (1,) + task.obs_space.shape  # Reformat to (channels, height, width)
device = "cuda"
config = {
    "device": device,
    "warm_up": 500,
    "update_freq": 50,
    "number_updates": 2,
    "batch_size": 300,
    "lr": 2e-4,
    "n_steps": 3,
    "tau": 0.01,
    "max_grad_norm": 10.0,
    "network_fn": lambda: network_fn(input_obs_shape, task.action_size, device),
}


# obs_space = (1,) + task.obs_space.shape
agent = DQNAgent(task.obs_space, task.action_space, **config)
env_runner = EnvRunner(task, agent, max_iterations=2000, data_logger=data_logger)

scores = env_runner.run(reward_goal=2, max_episodes=50000, gif_every_episodes=1000, force_new=True)
env_runner.interact_episode(render=True)
data_logger.close()


avg_length = 100
ma = running_mean(scores, avg_length)
# plot the scores
fig = plt.figure()
ax = fig.add_subplot(111)
plt.plot(range(len(scores)), scores)
plt.plot(range(avg_length, avg_length + len(ma)), ma)
plt.ylabel("Score")
plt.xlabel("Episode #")
plt.savefig(f"{env_name}.png", dpi=120)
plt.show()



from typing import Dict, List, Sequence, Union

Numeric = Union[int, float]
ObsType = ObservationType = Union[List[int], List[float]]
StateType = Union[int, List[float]]
ActionType = Union[int, float, List]
DoneType = bool
RewardType = Union[int, float]

HyperparameterType = Dict[str, str]
FeatureType = Sequence[int]
